# -*- coding: utf8 -*-
"""
.. module:: burpui.api.misc
    :platform: Unix
    :synopsis: Burp-UI misc api module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>

"""
from . import api, cache_key
from ..server import BUIServer  # noqa
from .custom import fields, Resource
from ..exceptions import BUIserverException
from ..ext.cache import cache

from six import iteritems
from flask import flash, url_for, current_app, session

import random
import re

bui = current_app  # type: BUIServer
ns = api.namespace('misc', 'Misc methods')

counters_fields = ns.model('Counters', {
    'phase': fields.Integer(description='Backup phase'),
    'Total': fields.List(fields.Integer, description='new/deleted/scanned/unchanged/total', attribute='total'),
    'Files': fields.List(fields.Integer, description='new/deleted/scanned/unchanged/total', attribute='files'),
    'Files (encrypted)': fields.List(fields.Integer, description='new/deleted/scanned/unchanged/total', attribute='files_encrypted'),
    'Meta data': fields.List(fields.Integer, description='new/deleted/scanned/unchanged/total', attribute='meta_data'),
    'Meta data (enc)': fields.List(fields.Integer, description='new/deleted/scanned/unchanged/total', attribute='meta_data_encrypted'),
    'Directories': fields.List(fields.Integer, description='new/deleted/scanned/unchanged/total', attribute='directories'),
    'Soft links': fields.List(fields.Integer, description='new/deleted/scanned/unchanged/total', attribute='soft_links'),
    'Hard links': fields.List(fields.Integer, description='new/deleted/scanned/unchanged/total', attribute='hard_links'),
    'Special files': fields.List(fields.Integer, description='new/deleted/scanned/unchanged/total', attribute='special_files'),
    'VSS headers': fields.List(fields.Integer, description='new/deleted/scanned/unchanged/total', attribute='vss_headers'),
    'VSS headers (enc)': fields.List(fields.Integer, description='new/deleted/scanned/unchanged/total', attribute='vss_headers_encrypted'),
    'VSS footers': fields.List(fields.Integer, description='new/deleted/scanned/unchanged/total', attribute='vss_footers'),
    'VSS footers (enc)': fields.List(fields.Integer, description='new/deleted/scanned/unchanged/total', attribute='vss_footers_encrypted'),
    'Grand total': fields.List(fields.Integer, description='new/deleted/scanned/unchanged/total', attribute='grand_total'),
    'warning': fields.Integer(description='Number of warnings so far'),
    'estimated_bytes': fields.Integer(description='Estimated Bytes in backup'),
    'bytes': fields.Integer(description='Bytes in backup'),
    'bytes_in': fields.Integer(description='Bytes received since backup started'),
    'bytes_out': fields.Integer(description='Bytes sent since backup started'),
    'start': fields.String(description='Timestamp of the start date of the backup'),
    'speed': fields.Integer(description='Backup speed', default=-1),
    'timeleft': fields.Integer(description='Estimated time left'),
    'percent': fields.Integer(required=True, description='Percentage done'),
    'path': fields.String(description='File that is currently treated by burp'),
})


@ns.route('/counters',
          '/<server>/counters',
          '/counters/<client>',
          '/<server>/counters/<client>',
          endpoint='counters')
@ns.doc(
    params={
        'server': 'Which server to collect data from when in multi-agent mode',
        'client': 'Client name',
    },
)
class Counters(Resource):
    """The :class:`burpui.api.misc.Counters` resource allows you to
    render the *live view* template of a given client.

    This resource is part of the :mod:`burpui.api.api` module.

    An optional ``GET`` parameter called ``serverName`` is supported when running
    in multi-agent mode.
    A mandatory ``GET`` parameter called ``clientName`` is used to know what client we
    are working on.
    """
    parser = ns.parser()
    parser.add_argument('serverName', help='Which server to collect data from when in multi-agent mode')
    parser.add_argument('clientName', help='Client name')
    monitor_fields = ns.model('Monitor', {
        'client': fields.String(required=True, description='Client name'),
        'agent': fields.String(description='Server (agent) name'),
        'counters': fields.Nested(counters_fields, description='Various statistics about the running backup'),
    })

    @ns.marshal_with(monitor_fields, code=200, description='Success')
    @ns.expect(parser)
    @ns.doc(
        responses={
            400: 'Missing argument',
            403: 'Insufficient permissions',
            404: 'Client not found in the running clients list',
        },
    )
    def get(self, server=None, client=None):
        """Returns counters for a given client

        **GET** method provided by the webservice.

        :param name: the client name if any. You can also use the GET parameter
        'name' to achieve the same thing

        :returns: Counters
        """
        args = self.parser.parse_args()
        server = server or args['serverName']
        client = client or args['clientName']
        # Check params
        if not client:
            self.abort(400, 'No client name provided')
        # Manage ACL
        if (bui.acl and
            not (bui.acl.is_client_allowed(self.username, client, server) or
                 self.is_admin)):
            self.abort(403, "Not allowed to view '{}' counters".format(client))
        bui.cli.is_one_backup_running()
        if isinstance(bui.cli.running, dict):
            if server and client not in bui.cli.running[server]:
                self.abort(404, "'{}' not found in the list of running clients for '{}'".format(client, server))
            else:
                found = False
                for (k, a) in iteritems(bui.cli.running):
                    found = found or (client in a)
                if not found:
                    api.bort(404, "'{}' not found in running clients".format(client))
        else:
            if client not in bui.cli.running:
                self.abort(404, "'{}' not found in running clients".format(client))
        try:
            counters = bui.cli.get_counters(client, agent=server)
        except BUIserverException:
            counters = {}
        res = {}
        res['client'] = client
        res['agent'] = server
        res['counters'] = counters
        return res


@ns.route('/monitor',
          '/<server>/monitor',
          endpoint='live')
@ns.doc(
    params={
        'server': 'Which server to collect data from when in multi-agent mode',
    },
)
class Live(Resource):
    """The :class:`burpui.api.misc.Live` resource allows you to
    retrieve a list of servers that are currently *alive*.

    This resource is part of the :mod:`burpui.api.misc` module.

    An optional ``GET`` parameter called ``serverName`` is supported when running
    in multi-agent mode.
    """
    parser = ns.parser()
    parser.add_argument('serverName', help='Which server to collect data from when in multi-agent mode')
    live_fields = ns.model('Live', {
        'client': fields.String(required=True, description='Client name'),
        'agent': fields.String(description='Server (agent) name'),
        'counters': fields.Nested(counters_fields, description='Various statistics about the running backup'),
    })

    @ns.marshal_list_with(live_fields, code=200, description='Success')
    @ns.expect(parser)
    def get(self, server=None):
        """Returns a list of clients that are currently running a backup

        **GET** method provided by the webservice.

        The *JSON* returned is:
        ::

            [
              {
                'client': 'client1',
                'agent': 'burp1',
                'counters': {
                    'phase': 2,
                    'path': '/etc/some/configuration',
                    '...': '...',
                },
              },
              {
                'client': 'client12',
                'agent': 'burp2',
                'counters': {
                    'phase': 3,
                    'path': '/etc/some/other/configuration',
                    '...': '...',
                },
              },
            ]

        The output is filtered by the :mod:`burpui.misc.acl` module so that you
        only see stats about the clients you are authorized to.

        :param server: Which server to collect data from when in multi-agent mode
        :type server: str

        :returns: The *JSON* described above
        """

        args = self.parser.parse_args()
        server = server or args['serverName']
        r = []
        # ACL
        if (bui.acl and
                not self.is_admin and
                server and
                server not in bui.acl.servers(self.username)):
            self.abort(403, 'You are not allowed to view stats of this server')
        if server:
            l = bui.cli.is_one_backup_running(server)[server]
            # ACL
            if bui.acl and not self.is_admin:
                allowed = bui.acl.clients(self.username, server)
                l = [x for x in l if x in allowed]
        else:
            l = bui.cli.is_one_backup_running()
        if isinstance(l, dict):
            for (k, a) in iteritems(l):
                for c in a:
                    # ACL
                    if (bui.acl and
                            not self.is_admin and
                            not bui.acl.is_client_allowed(
                                self.username,
                                c,
                                k)):
                        continue
                    s = {}
                    s['client'] = c
                    s['agent'] = k
                    try:
                        s['counters'] = bui.cli.get_counters(c, agent=k)
                    except BUIserverException:
                        s['counters'] = {}
                    r.append(s)
        else:
            for c in l:
                s = {}
                s['client'] = c
                try:
                    s['counters'] = bui.cli.get_counters(c, agent=server)
                except BUIserverException:
                    s['counters'] = {}
                r.append(s)
        return r


@ns.route('/alert', endpoint='alert')
class Alert(Resource):
    """The :class:`burpui.api.misc.Alert` resource allows you to propagate a
    message to the next screen.

    This resource is part of the :mod:`burpui.api.misc` module.
    """
    parser = ns.parser()
    parser.add_argument('message', required=True, help='Message to display')
    parser.add_argument('level', help='Alert level', choices=('danger', 'warning', 'info', 'success', '0', '1', '2', '3'), default='danger')

    @ns.expect(parser)
    @ns.doc(
        responses={
            201: 'Success',
        },
    )
    def post(self):
        """Propagate a message to the next screen (or whatever reads the session)"""
        def translate(level):
            levels = ['danger', 'warning', 'info', 'success']
            convert = {
                '0': 'success',
                '1': 'warning',
                '2': 'error',
                '3': 'info'
            }
            if not level:
                return 'danger'
            # return the converted value or the one we already had
            new = convert.get(level, level)
            # if the level is not handled, assume 'danger'
            if new not in levels:
                return 'danger'
            return new

        args = self.parser.parse_args()
        message = args['message']
        level = translate(args['level'])
        flash(message, level)
        return {'message': message, 'level': level}, 201


@ns.route('/about',
          '/<server>/about',
          endpoint='about')
@ns.doc(
    params={
        'server': 'Which server to collect data from when in multi-agent mode',
    },
)
class About(Resource):
    """The :class:`burpui.api.misc.About` resource allows you to retrieve
    various informations about ``Burp-UI``

    An optional ``GET`` parameter called ``serverName`` is supported when running
    in multi-agent mode.
    """
    api.LOGIN_NOT_REQUIRED.append('about')
    parser = ns.parser()
    parser.add_argument('serverName', help='Which server to collect data from when in multi-agent mode')
    burp_fields = ns.model('Burp', {
        'name': fields.String(required=True, description='Instance name', default='Burp'),
        'client': fields.String(description='Burp client version'),
        'server': fields.String(description='Burp server version'),
    })
    about_fields = ns.model('About', {
        'version': fields.String(required=True, description='Burp-UI version'),
        'release': fields.String(description='Burp-UI release (commit number)'),
        'api': fields.String(description='Burp-UI API documentation URL'),
        'burp': fields.Nested(burp_fields, as_list=True, description='Burp version'),
    })

    @cache.cached(timeout=3600, key_prefix=cache_key)
    @ns.marshal_with(about_fields, code=200, description='Success')
    @ns.expect(parser)
    def get(self, server=None):
        """Returns various informations about Burp-UI"""
        args = self.parser.parse_args()
        r = {}
        server = server or args['serverName']
        r['version'] = api.version
        r['release'] = api.release
        r['api'] = url_for('api.doc')
        r['burp'] = []
        cli = bui.cli.get_client_version(server)
        srv = bui.cli.get_server_version(server)
        multi = {}
        if isinstance(cli, dict):
            for (name, v) in iteritems(cli):
                multi[name] = {'client': v}
        if isinstance(srv, dict):
            for (name, v) in iteritems(srv):
                multi[name]['server'] = v
        if not multi:
            r['burp'].append({'client': cli, 'server': srv})
        else:
            for (name, v) in iteritems(multi):
                a = v
                a.update({'name': name})
                r['burp'].append(a)
        return r


@ns.route('/history',
          '/history/<client>',
          '/<server>/history',
          '/<server>/history/<client>',
          endpoint='history')
@ns.doc(
    params={
        'server': 'Which server to collect data from when in multi-agent mode',
        'client': 'Client name',
    },
)
class History(Resource):
    """The :class:`burpui.api.misc.History` resource allows you to retrieve
    an history of the backups

    An optional ``GET`` parameter called ``serverName`` is supported when
    running in multi-agent mode and ``clientName`` is also allowed to filter
    by client.

    ::

        $('#calendar').fullCalendar({

            eventSources: [

                // your event source
                {
                    events: [ // put the array in the `events` property
                        {
                            title  : 'event1',
                            start  : '2010-01-01'
                        },
                        {
                            title  : 'event2',
                            start  : '2010-01-05',
                            end    : '2010-01-07'
                        },
                        {
                            title  : 'event3',
                            start  : '2010-01-09T12:30:00',
                        }
                    ],
                    color: 'black',     // an option!
                    textColor: 'yellow' // an option!
                }

                // any other event sources...

            ]

        });

    """
    parser = ns.parser()
    parser.add_argument('serverName', help='Which server to collect data from when in multi-agent mode')
    parser.add_argument('clientName', help='Which client to collect data from')
    parser.add_argument('start', help='Return events after this date')
    parser.add_argument('end', help='Return events before this date')

    event_fields = ns.model('Event', {
        'title': fields.String(required=True, description='Event name'),
        'start': fields.DateTime(dt_format='iso8601', description='Start time of the event', attribute='date'),
        'end': fields.DateTime(dt_format='iso8601', description='End time of the event'),
        'name': fields.String(description='Client name'),
        'backup': fields.BackupNumber(description='Backup number', attribute='number'),
        'url': fields.String(description='Callback URL'),
    })
    history_fields = ns.model('History', {
        'events': fields.Nested(event_fields, as_list=True, description='Events list'),
        'color': fields.String(description='Background color'),
        'textColor': fields.String(description='Text color'),
        'name': fields.String(description='Feed name'),
    })

    @cache.cached(timeout=1800, key_prefix=cache_key)
    @ns.marshal_list_with(history_fields, code=200, description='Success')
    @ns.expect(parser)
    @ns.doc(
        responses={
            200: 'Success',
            403: 'Insufficient permissions',
        },
    )
    def get(self, client=None, server=None):
        """Returns a list of calendars describing the backups that have been
        completed so far

        **GET** method provided by the webservice.

        The *JSON* returned is:
        ::

            [
              {
                "color": "#7C6F44",
                "events": [
                  {
                    "backup": "0000001",
                    "end": "2015-01-25 13:32:04+01:00",
                    "name": "toto-test",
                    "start": "2015-01-25 13:32:00+01:00",
                    "title": "Client: toto-test, Backup n°0000001",
                    "url": "/client/toto-test"
                  }
                ],
                "name": "toto-test",
                "textColor": "white"
              }
            ]


        The output is filtered by the :mod:`burpui.misc.acl` module so that you
        only see stats about the clients you are authorized to.

        :param server: Which server to collect data from when in multi-agent mode
        :type server: str
        :param client: Which client to collect data from
        :type client: str

        :returns: The *JSON* described above
        """
        self._check_acl(client, server)
        return self._get_backup_history(client, server)

    def _check_acl(self, client=None, server=None):
        args = self.parser.parse_args()
        client = client or args['clientName']
        server = server or args['serverName']

        if (server and bui.acl and not self.is_admin and
                server not in bui.acl.servers(self.username)):
            self.abort(403, "You are not allowed to view this server infos")

        if (client and bui.acl and not self.is_admin and not
                bui.acl.is_client_allowed(self.username, client, server)):
            self.abort(403, "You are not allowed to view this client infos")

    def _get_backup_history(self, client=None, server=None, data=None):
        import arrow
        ret = []
        args = self.parser.parse_args()
        client = client or args['clientName']
        server = server or args['serverName']
        moments = {
            'start': None,
            'end': None
        }
        for moment in moments.keys():
            if moment in args:
                try:
                    if moment in args and args[moment] is not None:
                        moments[moment] = arrow.get(args[moment]).timestamp
                except arrow.parser.ParserError:
                    pass

        if client:
            (color, text) = self.gen_colors(client, server)
            feed = {
                'color': color,
                'textColor': text,
                'events': self.gen_events(client, moments, server, data),
            }
            name = client
            if server:
                name += ' on {}'.format(server)
            feed['name'] = name
            ret.append(feed)
            return ret
        elif server:
            if data and server in data:
                clients = [{'name': x} for x in data[server].keys()]
            else:
                clients = bui.cli.get_all_clients(agent=server)
            # manage ACL
            if bui.acl and not self.is_admin:
                clients = [x for x in clients if x['name'] in bui.acl.clients(self.username, server)]
            for cl in clients:
                (color, text) = self.gen_colors(cl['name'], server)
                feed = {
                    'events': self.gen_events(cl['name'], moments, server, data),
                    'textColor': text,
                    'color': color,
                    'name': '{} on {}'.format(cl['name'], server),
                }
                ret.append(feed)
            return ret

        if bui.standalone:
            if bui.acl and not self.is_admin:
                clients_list = bui.acl.clients(self.username)
            elif data:
                clients_list = data.keys()
            else:
                clients_list = [x['name'] for x in bui.cli.get_all_clients()]
            for cl in clients_list:
                (color, text) = self.gen_colors(cl)
                feed = {
                    'events': self.gen_events(cl, moments, data=data),
                    'textColor': text,
                    'color': color,
                    'name': cl,
                }
                ret.append(feed)
            return ret
        else:
            grants = {}
            if bui.acl and not self.is_admin:
                for serv in bui.acl.servers(self.username):
                    grants[serv] = bui.acl.clients(self.username, serv)
            else:
                for serv in bui.cli.servers:
                    grants[serv] = 'all'
            for (serv, clients) in iteritems(grants):
                if not isinstance(clients, list):
                    if data and serv in data:
                        clients = data[serv].keys()
                    else:
                        clients = [x['name'] for x in bui.cli.get_all_clients(agent=serv)]
                for cl in clients:
                    (color, text) = self.gen_colors(cl, serv)
                    feed = {
                        'events': self.gen_events(cl, moments, serv, data),
                        'textColor': text,
                        'color': color,
                        'name': '{} on {}'.format(cl, serv),
                    }
                    ret.append(feed)

        return ret

    def gen_colors(self, client=None, agent=None):
        """Generates color for an events feed"""
        cache = self._get_color_session(client, agent)
        if cache:
            return (cache['color'], cache['text'])
        labels = bui.cli.get_client_labels(client, agent)
        HTML_COLOR = r'((?P<hex>#(?P<red_hex>[0-9a-f]{1,2})(?P<green_hex>[0-9a-f]{1,2})(?P<blue_hex>[0-9a-f]{1,2}))|(?P<rgb>rgb\s*\(\s*(?P<red>2[0-5]{2}|2[0-4]\d|[0-1]?\d\d?)\s*,\s*(?P<green>2[0-5]{2}|2[0-4]\d|[0-1]?\d\d?)\s*,\s*(?P<blue>2[0-5]{2}|2[0-4]\d|[0-1]?\d\d?)\s*\))|(?P<plain>[\w-]+$))'
        color_found = False
        color = None
        text = None
        for label in labels:
            # We are looking for labels starting with "color:" or "text:"
            if re.search(r'^color:', label, re.IGNORECASE):
                search = re.search(r'^color:\s*{}'.format(HTML_COLOR), label, re.IGNORECASE)
                # we allow various color forms. For instance:
                # hex: #fa12e6
                # rgb: rgb (123, 42, 9)
                # plain: black
                if search.group('hex'):
                    red = search.group('red_hex')
                    green = search.group('green_hex')
                    blue = search.group('blue_hex')
                    # ensure ensure the hex part is of the form XX
                    red = red + red if len(red) == 1 else red
                    green = green + green if len(green) == 1 else green
                    blue = blue + blue if len(blue) == 1 else blue
                    # Now convert the hex to an int
                    red = int(red, 16)
                    green = int(green, 16)
                    blue = int(blue, 16)
                elif search.group('rgb'):
                    red = int(search.group('red'))
                    green = int(search.group('green'))
                    blue = int(search.group('blue'))
                elif search.group('plain'):
                    # if plain color is provided, we cannot guess the adapted
                    # text color, so we assume white (unless text is specified)
                    red = 0
                    green = 0
                    blue = 0
                    color = search.group('plain')
                else:
                    continue
                color = color or '#{:02X}{:02X}{:02X}'.format(red, green, blue)
                color_found = True
            if re.search(r'^text:', label, re.IGNORECASE):
                search = re.search(r'^text:\s*{}'.format(HTML_COLOR), label, re.IGNORECASE)
                # if we don't find anything, we'll generate a color based on
                # the value of the red, green and blue variables
                text = search.group('hex') or search.group('rgb') or search.group('plain')
            if color and text:
                break

        if not color_found:
            def rand():
                return random.randint(0, 255)
            red = rand()
            green = rand()
            blue = rand()

        text = text or self._get_text_color(red, green, blue)
        color = color or '#{:02X}{:02X}{:02X}'.format(red, green, blue)
        self._set_color_session(color, text, client, agent)
        return (color, text)

    def _get_text_color(self, red=0, green=0, blue=0):
        """Generates the text color for a given color"""
        yiq = ((red * 299) + (green * 587) + (blue * 114)) / 1000
        return 'black' if yiq >= 128 else 'white'

    def _get_color_session(self, client, agent=None):
        """Since we can *paginate* the rendering, we need to store the already
        generated colors

        This method allows to retrieve already generated colors if any
        """
        sess = session._get_current_object()
        if 'colors' in sess:
            colors = sess['colors']
            if agent and agent in colors:
                return colors[agent].get(client)
            elif not agent:
                return colors.get(client)
        return None

    def _set_color_session(self, color, text, client, agent=None):
        """Since we can *paginate* the rendering, we need to store the already
        generated colors

        This method allows to store already generated colors in the session
        """
        sess = session._get_current_object()
        dic = {}
        if agent:
            if 'colors' in sess and agent in sess['colors']:
                dic[agent] = sess['colors'][agent]
            else:
                dic[agent] = {}
            dic[agent][client] = {
                'color': color,
                'text': text
            }
        else:
            dic[client] = {
                'color': color,
                'text': text
            }
        if 'colors' in sess:
            sess['colors'].update(dic)
        else:
            sess['colors'] = dic

    def gen_events(self, client, moments, server=None, data=None):
        """Creates events for a given client"""
        events = []
        filtered = False
        if data:
            if bui.standalone:
                events = data.get(client)
            else:
                events = data.get(server, {}).get(client)
        if not events:
            events = bui.cli.get_client_filtered(client, start=moments['start'], end=moments['end'], agent=server)
            filtered = True

        ret = []
        for ev in events:
            if data and not filtered:
                # events are sorted by date DESC
                if moments['start'] and ev['date'] < moments['start']:
                    continue
                if moments['end'] and ev['date'] > moments['end']:
                    continue
            ev['title'] = 'Client: {0}, Backup n°{1:07d}'.format(client, int(ev['number']))
            if server:
                ev['title'] += ', Server: {0}'.format(server)
            ev['name'] = client
            ev['url'] = url_for('view.backup_report', name=client, server=server, backup=int(ev['number']))
            ret.append(ev)

        return ret



# -*- coding: utf8 -*-
"""
.. module:: burpui.misc.parser.openssl
    :platform: Unix
    :synopsis: Burp-UI configuration file parser OpenSSL configuration.
.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>
"""
import os
import re
import codecs
import logging
import subprocess

from ..._compat import PY3

from hashlib import md5
from six import iteritems
from OpenSSL import crypto

if PY3:
    long = int


class OSSLAuth(object):
    """OpenSSL wrapper"""

    def __init__(self, ca_name, ossl_conf, global_conf):
        """
        :param ca_name: CA name
        :type ca_name: str

        :param ossl_conf: OpenSSL config
        :type ossl_conf: :class:`burpui.misc.parser.openssl.OSSLConf`

        :param global_conf: Global config
        :type global_conf: :class:`burpui.misc.parser.utils.Config`
        """
        self.logger = logging.getLogger('burp-ui')
        self.name = ca_name
        self.ossl_conf = ossl_conf
        self.global_conf = global_conf
        self._load_crl()

    def _load_crl(self):
        crl_path = self._get_crl_path()
        try:
            with open(crl_path) as crl:
                self.crl = crypto.load_crl(
                    crypto.FILETYPE_PEM,
                    crl.read()
                )
        except IOError as err:
            self.logger.warning(str(err))
            self.crl = None

    def _get_crt_path(self, client):
        return '{}.crt'.format(
            os.path.join(self.ossl_conf.values.get('CA_DIR'), client)
        )

    def _get_crl_path(self):
        """Returns the CRL path of a given CA"""
        path = self.ossl_conf.values.get('CA_DIR')
        if not path:
            return ''
        return '{}/CA_{}.crl'.format(path, self.name)

    def check_client_revoked(self, client):
        """Check whether the given client certificate has been revoked
        :param client: Client name
        :type client: str

        :returns: True or False
        """
        if not self.crl:
            return False

        c_cert = self._get_crt_path(client)
        try:
            with open(c_cert) as crt:
                client_crt = crypto.load_certificate(
                    crypto.FILETYPE_PEM,
                    crt.read()
                )
        except IOError as err:
            self.logger.warning(str(err))
            return False

        for rvk in self.crl.get_revoked():
            if client_crt.get_serial_number() == long(rvk.get_serial(), 16):
                return True

        return False

    def revoke_client(self, client, update=True):
        """Revoke a given client

        :param client: Client name
        :type client: str

        :param update: Whether to update the CRL or not
        :type update: bool

        :returns: True or False
        """
        if not client:
            return False
        c_cert = self._get_crt_path(client)
        try:
            DEVNULL = open(os.devnull, 'w')
            openssl = subprocess.check_output([
                'which',
                'openssl'],
                stderr=DEVNULL
            ).rstrip('\n')
            # FIXME: Don't know why pyOpenssl does not return the hex serial :-/
            _, serial = subprocess.check_output([
                openssl,
                'x509',
                '-serial',
                '-noout',
                '-in',
                c_cert],
                stderr=DEVNULL
            ).rstrip('\n').split('=')
            self.logger.debug('{} serial: {}'.format(client, serial))
            subprocess.check_call([
                self.global_conf.get('ca_burp_ca'),
                '--name',
                self.name,
                '--config',
                self.global_conf.get('ca_conf'),
                '--dir',
                self.ossl_conf.values.get('CA_DIR'),
                '--revoke',
                serial],
                stderr=subprocess.STDOUT,
                stdout=DEVNULL
            )
            if update:
                subprocess.check_call([
                    self.global_conf.get('ca_burp_ca'),
                    '--name',
                    self.name,
                    '--config',
                    self.global_conf.get('ca_conf'),
                    '--dir',
                    self.ossl_conf.values.get('CA_DIR'),
                    '--crl'],
                    stderr=subprocess.STDOUT,
                    stdout=DEVNULL
                )
                self._load_crl()
        except subprocess.CalledProcessError as err:
            self.logger.warning(str(err))
            return False
        return True


class OSSLConf(object):
    """Parse the given OpenSSL configuration file

    :param conffile: Configuration file to parse
    :type conffile: str
    """

    def __init__(self, conffile=None):
        self.conffile = conffile
        self.conf = {}
        self.env_cache = {}
        self.md5 = None

    @property
    def values(self):
        """Mapping of the configuration file into a dict"""
        self._parse()
        return self.conf

    def _read(self):
        """Read the file if possible"""
        ret = []
        try:
            with codecs.open(self.conffile, 'r', 'utf-8') as fil:
                ret = [x.rstrip('\n') for x in fil.readlines()]
        except IOError:
            pass
        return ret

    def _parse(self):
        """Parse the file if it has changed since last read and replace all
        variables using a multi-passes loop
        """
        dic = {}
        chksum = self._md5(self.conffile)
        if self.md5 and self.md5 == chksum:
            return

        for line in self._read():
            if re.match(r'^\s*#', line):
                continue
            res = re.search(r'\s*([^#][^=\s]+)\s*=\s*(.*)$', line)
            if res:
                key = res.group(1)
                val = res.group(2)

                dic[key] = val

        for key, val in iteritems(dic):
            self.conf[key] = dic[key] = val = self._translate(dic, key, val)

        self.md5 = chksum

    def _translate(self, temp, key, val):
        """Translate variables if needed until the returned value does not seem
        to contain variables anymore
        """
        res = val
        dic = temp
        env = self._is_env(res)
        if env:
            for match in env:
                res = dic[key] = self._translate_env(
                    dic,
                    match,
                    res
                )
        lcl = self._is_local(res)
        if lcl:
            for match in lcl:
                res = dic[key] = self._translate_local(
                    dic,
                    match,
                    res
                )
        if not lcl and not env:
            return res
        return self._translate(dic, key, res)

    @staticmethod
    def _is_local(val):
        """Look for 'local' variables (ie. $dir)"""
        lcl = re.compile(r'\${?(\w+)}?')
        res = [x for x in lcl.findall(val) if x.lower() != 'env']
        return res if res else False

    def _translate_local(self, temp, pattern, val):
        """Translate 'local' variables (ie. $dir is replaced by the content of
        the 'dir' variable)
        """
        res = self.conf.get(pattern) or temp.get(pattern) or ''
        return re.sub(r'\${?' + pattern + '}?', res, val)

    @staticmethod
    def _is_env(val):
        """Look for 'global' variables (ie. $ENV::HOME)"""
        env = re.compile(r'\${?ENV::(\w+)}?')
        res = env.findall(val)
        return res if res else False

    def _translate_env(self, temp, pattern, val):
        """Translate 'global' variables (ie. $ENV::HOME is replace by the
        content of the 'HOME' variable if available in the file or with the
        $HOME environment variable
        """
        res = self.env_cache.get(pattern) or os.getenv(pattern) or \
            self.conf.get(pattern) or temp.get(pattern) or ''
        self.env_cache[pattern] = res
        return re.sub(r'\${?ENV::' + pattern + '}?', res, val)

    @staticmethod
    def _md5(path):
        """Compute the md5sum of the configuration file to detect changes"""
        hash_md5 = md5()
        try:
            with open(path, "rb") as bfile:
                for chunk in iter(lambda: bfile.read(4096), b""):
                    hash_md5.update(chunk)
            return hash_md5.hexdigest()
        except IOError:
            return None



# -*- coding: utf8 -*-
from .interface import BUIacl, BUIaclLoader

import json


class ACLloader(BUIaclLoader):
    """See :class:`burpui.misc.acl.interface.BUIaclLoader`"""
    def __init__(self, app=None):
        """See :func:`burpui.misc.acl.interface.BUIaclLoader.__init__`

        :param app: Application context
        :type app: :class:`burpui.server.BUIServer`
        """
        self.app = app
        self.admins = [
            'admin'
        ]
        self.clients = {}
        self.servers = {}
        self.standalone = self.app.standalone
        conf = self.app.conf
        adms = []
        if 'BASIC:ACL' in conf.options:
            adms = conf.safe_get('admin', 'force_list', section='BASIC:ACL')
            for opt in conf.options.get('BASIC:ACL').keys():
                if opt == 'admin':
                    continue
                lit = conf.safe_get(opt, section='BASIC:ACL')
                rec = []
                try:
                    rec = json.loads(lit)
                    if isinstance(rec, dict):
                        self.servers[opt] = rec.keys()
                except Exception as e:
                    self.logger.error(str(e))
                    rec = [lit]
                self.clients[opt] = rec

        if adms and adms != [None]:
            self.admins = adms
        self._acl = BasicACL(self)
        self.logger.debug('admins: ' + str(self.admins))
        self.logger.debug('clients: ' + str(self.clients))
        self.logger.debug('servers: ' + str(self.servers))

    @property
    def acl(self):
        """Property to retrieve the backend"""
        if self._acl:
            return self._acl
        return None  # pragma: no cover


class BasicACL(BUIacl):
    """See :class:`burpui.misc.acl.interface.BUIacl`"""
    def __init__(self, handler=None):
        """:func:`burpui.misc.acl.interface.BUIacl.__init__` instanciate ACL
        engine.

        :param handler: ACL handler
        :type handler: :class:`burpui.misc.acl.interface.BUIaclLoader`
        """
        if not handler:  # pragma: no cover
            return
        self.handler = handler
        self.standalone = handler.standalone
        self.admins = handler.admins
        self.cls = handler.clients
        self.srv = handler.servers

    def is_admin(self, username=None):
        """See :func:`burpui.misc.acl.interface.BUIacl.is_admin`"""
        if not username:  # pragma: no cover
            return False
        return username in self.admins

    def clients(self, username=None, server=None):
        """See :func:`burpui.misc.acl.interface.BUIacl.clients`"""
        if not username:  # pragma: no cover
            return []
        if username in self.cls:
            cls = self.cls[username]
            if server and isinstance(cls, dict):
                if server in cls:
                    return cls[server]
                return []
            # No server defined whereas we have an extended ACL
            if not server and self.servers(username):
                return []
            return cls
        return [username]

    def servers(self, username=None):
        """See :func:`burpui.misc.acl.interface.BUIacl.servers`"""
        if username and username in self.srv:
            return self.srv[username]
        return []

    def is_client_allowed(self, username=None, client=None, server=None):
        """See :func:`burpui.misc.acl.interface.BUIacl.is_client_allowed`"""
        if not username or not client:  # pragma: no cover
            return False
        # No server defined whereas we have an extended ACL
        if not server and self.servers(username):
            return False
        cls = self.clients(username, server)
        return (cls and client in cls) or self.is_admin(username)



# -*- coding: utf8 -*-
"""
.. module:: burpui.api.custom
    :platform: Unix
    :synopsis: Burp-UI api custom module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>

"""
from .resource import Resource  # noqa



# -*- coding: utf8 -*-
"""
.. module:: burpui.misc.acl.interface
    :platform: Unix
    :synopsis: Burp-UI ACL interface.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>

"""
from abc import ABCMeta, abstractmethod, abstractproperty
from six import with_metaclass

import logging


class BUIaclLoader(with_metaclass(ABCMeta, object)):
    """The :class:`burpui.misc.acl.interface.BUIaclLoader` class is used to
    load the actual ACL backend"""

    logger = logging.getLogger('burp-ui')

    def __init__(self, app=None):
        """:func:`burpui.misc.acl.interface.BUIaclLoader.__init__` instanciate
        the loader.

        :param app: Application context
        :type app: :class:`burpui.server.BUIServer`
        """
        pass  # pragma: no cover

    @abstractproperty
    @property
    def acl(self):
        """Property to retrieve the backend"""
        return None  # pragma: no cover


class BUIacl(with_metaclass(ABCMeta, object)):
    """The :class:`burpui.misc.acl.interface.BUIacl` class represents the ACL
    engine.
    """

    @abstractmethod
    def is_admin(self, username=None):
        """:func:`burpui.misc.acl.interface.BUIacl.is_admin` is used to know if
        a user has administrator rights.

        :param username: Username to check
        :type username: str

        :returns: True if the user has admin rights, otherwise False
        """
        return False  # pragma: no cover

    @abstractmethod
    def clients(self, username=None, server=None):
        """:func:`burpui.misc.acl.interface.BUIacl.clients` returns a list of
        allowed clients for a given user.

        :param username: Username to check
        :type username: str

        :param server: Server name. Used in multi-agent mode
        :type server: str

        :returns: A list of clients
        """
        return []  # pragma: no cover

    @abstractmethod
    def servers(self, username=None):
        """:func:`burpui.misc.acl.interface.BUIacl.servers` returns a list of
        allowed servers for a given user.

        :param username: Username to check
        :type username: str

        :returns: A list of servers
        """
        return []  # pragma: no cover

    @abstractmethod
    def is_client_allowed(self, username=None, client=None, server=None):
        """:func:`burpui.misc.acl.interface.BUIacl.is_client_allowed` tells us
        if a given user has access to a given client on a given server.

        :param username: Username to check
        :type username: str

        :param client: Client to check
        :type client: str

        :param server: Server to check
        :type server: str

        :returns: True if username is granted, otherwise False
        """
        return False  # pragma: no cover



#!/usr/bin/env python
# -*- coding: utf8 -*-
import sys
import os
import json
import unittest
import tempfile

if sys.version_info >= (3, 0):
    from urllib.request import urlopen
else:
    from urllib2 import urlopen

from flask_testing import LiveServerTestCase, TestCase
from flask import url_for, session

sys.path.append('{0}/..'.format(os.path.join(os.path.dirname(os.path.realpath(__file__)))))

from burpui import init as BUIinit


class BurpuiLiveTestCase(LiveServerTestCase):

    def create_app(self):
        conf = os.path.join(os.path.dirname(os.path.realpath(__file__)), '../share/burpui/etc/burpui.sample.cfg')
        bui = BUIinit(debug=12, gunicorn=False, unittest=True)
        bui.setup(conf)
        bui.config['DEBUG'] = False
        bui.config['TESTING'] = True
        bui.config['LOGIN_DISABLED'] = True
        bui.config['LIVESERVER_PORT'] = 5001
        bui.config['CFG'] = conf
        bui.login_manager.init_app(bui)
        return bui

    def setUp(self):
        print ('\nBegin Test 1\n')

    def tearDown(self):
        print ('\nTest 1 Finished!\n')

    def test_server_is_up_and_running(self):
        import socket
        import errno
        try:
            response = urlopen(self.get_server_url())
            self.assertEqual(response.code, 200)
        except socket.error as exp:
            if exp.errno != errno.ECONNRESET:
                raise
            pass


class BurpuiAPIBasicHTTPTestCase(TestCase):

    def setUp(self):
        print ('\nBegin Test 2\n')

    def tearDown(self):
        print ('\nTest 2 Finished!\n')
        os.unlink(self.logfile)

    def create_app(self):
        conf = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'test2.cfg')
        _, self.logfile = tempfile.mkstemp()
        bui = BUIinit(conf, 1, self.logfile, gunicorn=False, unittest=True)
        bui.config['DEBUG'] = False
        return bui

    def test_auth_required(self):
        response = self.client.get(url_for('api.about'))
        self.assert200(response)
        response = self.client.get(url_for('api.counters'))
        self.assert401(response)

    def test_auth_valid(self):
        import base64
        response = self.client.get(
            url_for('api.live'),
            headers={
                'Authorization': 'Basic ' + base64.b64encode(b'admin:admin').decode('utf-8')
            }
        )
        self.assert200(response)


class BurpuiAPITestCase(TestCase):

    def setUp(self):
        print ('\nBegin Test 3\n')

    def tearDown(self):
        print ('\nTest 3 Finished!\n')

    def create_app(self):
        conf = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'test2.cfg')
        bui = BUIinit(gunicorn=False, unittest=True)
        bui.setup(conf)
        bui.config['TESTING'] = True
        bui.config['LOGIN_DISABLED'] = True
        bui.config['CFG'] = conf
        bui.login_manager.init_app(bui)
        self.bui = bui
        return bui

    def test_no_clients(self):
        response = self.client.get(url_for('api.clients_stats'))
        self.assertEquals(json.loads(response.data.decode('utf-8'))['message'], u'Cannot contact burp server at 127.0.0.1:9999')
        self.assert500(response)

    def test_server_config_parsing(self):
        response = self.client.get(url_for('api.server_settings'))
        asse = dict((
                    (
                        u'results',
                        {
                            u'common': [],
                            u'boolean': [],
                            u'integer': [],
                            u'multi': [],
                            u'includes': [],
                            u'includes_ext': [],
                            u'clients': []
                        }
                    ),
                    (u'boolean', self.bui.cli.get_parser_attr('boolean_srv')),
                    (u'string', self.bui.cli.get_parser_attr('string_srv')),
                    (u'integer', self.bui.cli.get_parser_attr('integer_srv')),
                    (u'multi', self.bui.cli.get_parser_attr('multi_srv')),
                    (u'server_doc', self.bui.cli.get_parser_attr('doc')),
                    (u'suggest', self.bui.cli.get_parser_attr('values')),
                    (u'placeholders', self.bui.cli.get_parser_attr('placeholders')),
                    (u'defaults', self.bui.cli.get_parser_attr('defaults'))))
        self.assertEquals(response.json, asse)

    def test_client_config_parsing(self):
        response = self.client.get(url_for('api.client_settings', client='toto'))
        asse = dict((
                    (
                        u'results',
                        {
                            u'common': [],
                            u'boolean': [],
                            u'integer': [],
                            u'multi': [],
                            u'includes': [],
                            u'includes_ext': [],
                            u'clients': []
                        }
                    ),
                    (u'boolean', self.bui.cli.get_parser_attr('boolean_cli')),
                    (u'string', self.bui.cli.get_parser_attr('string_cli')),
                    (u'integer', self.bui.cli.get_parser_attr('integer_cli')),
                    (u'multi', self.bui.cli.get_parser_attr('multi_cli')),
                    (u'server_doc', self.bui.cli.get_parser_attr('doc')),
                    (u'suggest', self.bui.cli.get_parser_attr('values')),
                    (u'placeholders', self.bui.cli.get_parser_attr('placeholders')),
                    (u'defaults', self.bui.cli.get_parser_attr('defaults'))))
        self.assertEquals(response.json, asse)

    def test_restore(self):
        response = self.client.post(url_for('api.restore', name='dummy', backup=1), data=dict(strip=False))
        self.assert400(response)

    def test_running_clients(self):
        response = self.client.get(url_for('api.running_clients'))
        self.assertEquals(response.json, [])

    def test_live_rendering(self):
        response = self.client.get(url_for('api.counters', client='toto'))
        self.assert404(response)
        response = self.client.get(url_for('api.counters'))
        self.assert400(response)

    def test_servers_json(self):
        response = self.client.get(url_for('api.servers_stats'))
        self.assertEquals(response.json, [])

    def test_live(self):
        response = self.client.get(url_for('api.live'))
        self.assertEquals(response.json, [])

    def test_running(self):
        response = self.client.get(url_for('api.running_backup'))
        self.assertEquals(response.json, dict(running=False))

    def test_client_tree(self):
        response = self.client.get(url_for('api.client_tree', name='toto', backup=1))
        self.assertEquals(json.loads(response.data.decode('utf-8'))['message'], u'Cannot contact burp server at 127.0.0.1:9999')
        self.assert500(response)

    def test_clients_report_json(self):
        response = self.client.get(url_for('api.clients_report'))
        self.assertEquals(json.loads(response.data.decode('utf-8'))['message'], u'Cannot contact burp server at 127.0.0.1:9999')
        self.assert500(response)

    def test_client_stat_json(self):
        response = self.client.get(url_for('api.client_stats', name='toto'))
        self.assertEquals(json.loads(response.data.decode('utf-8'))['message'], u'Cannot contact burp server at 127.0.0.1:9999')
        self.assert500(response)
        response = self.client.get(url_for('api.client_stats', name='toto', backup=1))
        self.assertEquals(json.loads(response.data.decode('utf-8'))['message'], u'Cannot contact burp server at 127.0.0.1:9999')
        self.assert500(response)

    def test_client_json(self):
        response = self.client.get(url_for('api.client_report', name='toto'))
        self.assertEquals(json.loads(response.data.decode('utf-8'))['message'], u'Cannot contact burp server at 127.0.0.1:9999')
        self.assert500(response)


class BurpuiRoutesTestCase(TestCase):

    def setUp(self):
        print ('\nBegin Test 4\n')

    def tearDown(self):
        print ('\nTest 4 Finished!\n')

    def create_app(self):
        conf = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'test4.cfg')
        bui = BUIinit(conf, gunicorn=False, unittest=True)
        bui.setup(conf)
        bui.config['TESTING'] = True
        bui.config['LOGIN_DISABLED'] = True
        bui.config['LIVESERVER_PORT'] = 5001
        bui.config['SECRET_KEY'] = 'toto'
        bui.login_manager.init_app(bui)
        return bui

    def test_live_monitor(self):
        response = self.client.get(url_for('view.live_monitor'), follow_redirects=True)
        assert 'Sorry, there are no running backups' in response.data.decode('utf-8')

    def test_get_clients(self):
        response = self.client.get(url_for('api.clients_stats'))
        self.assertEqual(response.json, [{u'state': u'idle', u'last': u'never', u'human': u'never', u'name': u'testclient', u'phase': None, u'percent': 0}])


class BurpuiLoginTestCase(TestCase):

    def setUp(self):
        print ('\nBegin Test 5\n')

    def tearDown(self):
        print ('\nTest 5 Finished!\n')

    def login(self, username, password):
        return self.client.post(url_for('view.login'), data=dict(
            username=username,
            password=password
        ), follow_redirects=True)

    def create_app(self):
        conf = os.path.join(os.path.dirname(os.path.realpath(__file__)), '../share/burpui/etc/burpui.sample.cfg')
        bui = BUIinit(conf, False, None, False, unittest=True)
        bui.config['TESTING'] = True
        bui.config['LIVESERVER_PORT'] = 5001
        bui.config['WTF_CSRF_ENABLED'] = False
        bui.cli.port = 9999
        return bui

    def test_config_render(self):
        rv = self.login('admin', 'admin')
        response = self.client.get(url_for('view.settings'))
        assert 'Burp Configuration' in response.data.decode('utf-8')

    def test_login_ok(self):
        rv = self.login('admin', 'admin')
        assert 'Logged in successfully' in rv.data.decode('utf-8')

    def test_login_ko(self):
        rv = self.login('admin', 'toto')
        assert 'Wrong username or password' in rv.data.decode('utf-8')

    def test_login_no_user(self):
        rv = self.login('toto', 'toto')
        assert 'Wrong username or password' in rv.data.decode('utf-8')


class BurpuiACLTestCase(TestCase):

    def setUp(self):
        print ('\nBegin Test 6\n')

    def tearDown(self):
        print ('\nTest 6 Finished!\n')

    def login(self, username, password):
        return self.client.post(url_for('view.login'), data=dict(
            username=username,
            password=password
        ), follow_redirects=True)

    def logout(self):
        return self.client.get(url_for('view.logout'), follow_redirects=True)

    def create_app(self):
        conf = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'test6.cfg')
        bui = BUIinit(conf, False, None, False, unittest=True)
        bui.config['TESTING'] = True
        bui.config['LIVESERVER_PORT'] = 5001
        bui.config['WTF_CSRF_ENABLED'] = False
        bui.cli.port = 9999
        return bui

    def test_login_ko(self):
        with self.client:
            rv = self.login('admin', 'toto')
            assert 'Wrong username or password' in rv.data.decode('utf-8')
            self.logout()

    def test_config_render(self):
        with self.client:
            rv = self.login('admin', 'admin')
            response = self.client.get(url_for('view.settings'))
            assert 'Burp Configuration' in response.data.decode('utf-8')
            self.logout()

    def test_admin_api(self):
        with self.client:
            rv = self.login('admin', 'admin')
            response = self.client.get(url_for('api.auth_users'))
            response2 = self.client.get(url_for('api.auth_backends'))
            self.assertEqual(response.json, [{u'id': u'admin', u'name': u'admin', u'backend': u'BASIC'}, {u'id': u'user1', u'name': u'user1', u'backend': u'BASIC'}])
            self.assertEqual(response2.json, [{u'add': True, u'del': True, u'name': u'BASIC', u'mod': True}])

    def test_config_render_ko(self):
        with self.client:
            rv = self.login('user1', 'password')
            response = self.client.get(url_for('view.settings'))
            self.assert403(response)
            self.logout()

    def test_cli_settings_ko(self):
        with self.client:
            rv = self.login('user1', 'password')
            response = self.client.get(url_for('api.client_settings', client='toto'))
            self.assert403(response)
            self.logout()


class BurpuiTestInit(TestCase):

    def setUp(self):
        print ('\nBegin Test 7\n')

    def tearDown(self):
        print ('\nTest 7 Finished!\n')
        os.unlink(self.tmpFile)

    def create_app(self):
        conf1 = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'test7-1.cfg')
        conf2 = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'test7-2.cfg')
        BUIinit(conf1, False, None, False, unittest=True)
        BUIinit(conf2, False, None, False, unittest=True)
        bui = BUIinit(None, False, None, False, unittest=True)
        bui.config['TESTING'] = True
        bui.config['LIVESERVER_PORT'] = 5001
        bui.config['WTF_CSRF_ENABLED'] = False
        bui.cli.port = 9999
        return bui

    def test_exception(self):
        _, self.tmpFile = tempfile.mkstemp()
        self.assertRaises(IOError, BUIinit, 'thisfileisnotlikelytoexist', True, self.tmpFile, False, unittest=True)
        self.assertRaises(IOError, BUIinit, 'thisfileisnotlikelytoexist', False, self.tmpFile, False, unittest=True)
        conf3 = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'test7-3.cfg')
        self.assertRaises(ImportError, BUIinit, conf3, False, None, False, unittest=True)


#class BurpuiAPILoginTestCase(TestCase):
#
#    def setUp(self):
#        print ('\nBegin Test 7\n')
#
#    def tearDown(self):
#        print ('\nTest 7 Finished!\n')
#
#    def login(self, username, password):
#        return self.client.post(url_for('view.login'), data=dict(
#            username=username,
#            password=password
#        ), follow_redirects=True)
#
#    def create_app(self):
#        conf = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'test7.cfg')
#        app.config['TESTING'] = True
#        app.config['LOGIN_DISABLED'] = True
#        app.config['CFG'] = conf
#        bui.setup(conf)
#        return app
#
#    def test_server_config_parsing(self):
#        rv = self.login('toto', 'toto')
#        response = self.client.get(url_for('api.server_settings', server='dummy'))
#        self.assertEquals(response.json, {u'message': u'Sorry, you don\'t have rights to access the setting panel'})
#
#    def test_client_config_parsing(self):
#        rv = self.login('toto', 'toto')
#        response = self.client.get(url_for('api.client_settings', client='toto', server='dummy'))
#        self.assertEquals(response.json, {u'message': u'Sorry, you don\'t have rights to access the setting panel'})
#
#    def test_restore(self):
#        rv = self.login('toto', 'toto')
#        response = self.client.post(url_for('api.restore', name='dummy', backup=1, server='dummy'), data=dict(strip=False))
#        self.assert500(response)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf8 -*-
"""
.. module:: burpui.api.clients
    :platform: Unix
    :synopsis: Burp-UI clients api module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>

"""
from . import api, cache_key
from ..server import BUIServer  # noqa
from .custom import fields, Resource
from ..exceptions import BUIserverException

from six import iteritems
from flask import current_app

bui = current_app  # type: BUIServer
ns = api.namespace('clients', 'Clients methods')


# Seem to not be used anymore
# TODO: we can probably remove this someday
@ns.route('/running',
          '/<server>/running',
          '/running/<client>',
          '/<server>/running/<client>',
          endpoint='running_clients')
@ns.doc(
    params={
        'server': 'Which server to collect data from when in multi-agent mode',
        'client': 'Client name',
    },
)
class RunningClients(Resource):
    """The :class:`burpui.api.clients.RunningClients` resource allows you to
    retrieve a list of clients that are currently running a backup.

    This resource is part of the :mod:`burpui.api.clients` module.

    An optional ``GET`` parameter called ``serverName`` is supported when running
    in multi-agent mode.
    """
    parser = ns.parser()
    parser.add_argument('serverName', help='Which server to collect data from when in multi-agent mode')

    @ns.expect(parser)
    def get(self, client=None, server=None):
        """Returns a list of clients currently running a backup

        **GET** method provided by the webservice.

        The *JSON* returned is:
        ::

            [ 'client1', 'client2' ]


        The output is filtered by the :mod:`burpui.misc.acl` module so that you
        only see stats about the clients you are authorized to.

        :param server: Which server to collect data from when in multi-agent mode
        :type server: str

        :param client: Ask a specific client in order to know if it is running a backup
        :type client: str

        :returns: The *JSON* described above.
        """
        server = server or self.parser.parse_args()['serverName']
        if client:
            if bui.acl:
                if (not self.is_admin and not
                        bui.acl.is_client_allowed(self.username,
                                                  client,
                                                  server)):
                    r = []
                    return r
            if bui.cli.is_backup_running(client, server):
                r = [client]
                return r
            else:
                r = []
                return r

        r = bui.cli.is_one_backup_running(server)
        # Manage ACL
        if bui.acl and not self.is_admin:
            if isinstance(r, dict):
                new = {}
                for serv in bui.acl.servers(self.username):
                    allowed = bui.acl.clients(self.username, serv)
                    new[serv] = [x for x in r[serv] if x in allowed]
                r = new
            else:
                allowed = bui.acl.clients(self.username, server)
                r = [x for x in r if x in allowed]
        return r


@ns.route('/backup-running',
          '/<server>/backup-running',
          endpoint='running_backup')
@ns.doc(
    params={
        'server': 'Which server to collect data from when in multi-agent mode',
    }
)
class RunningBackup(Resource):
    """The :class:`burpui.api.clients.RunningBackup` resource allows you to
    access the status of the server in order to know if there is a running
    backup currently.

    This resource is part of the :mod:`burpui.api.clients` module.
    """
    running_fields = ns.model('Running', {
        'running': fields.Boolean(required=True, description='Is there a backup running right now'),
    })

    @ns.marshal_with(running_fields, code=200, description='Success')
    def get(self, server=None):
        """Tells if a backup is running right now

        **GET** method provided by the webservice.

        The *JSON* returned is:
        ::

            {
                "running": false
            }


        The output is filtered by the :mod:`burpui.misc.acl` module so that you
        only see stats about the clients you are authorized to.

        :param server: Which server to collect data from when in multi-agent mode
        :type server: str

        :returns: The *JSON* described above.
        """
        return {
            'running': self._is_one_backup_running(
                bui.cli.is_one_backup_running(server),
                server
            )
        }

    def _is_one_backup_running(self, res, server):
        """Check if a backup is running"""
        # Manage ACL
        if bui.acl and not self.is_admin:
            if isinstance(res, dict):
                new = {}
                for serv in bui.acl.servers(self.username):
                    allowed = bui.acl.clients(self.username, serv)
                    new[serv] = [x for x in res[serv] if x in allowed]
                res = new
            else:
                allowed = bui.acl.clients(self.username, server)
                res = [x for x in res if x in allowed]
        running = False
        if isinstance(res, dict):
            for (_, run) in iteritems(res):
                running = running or (len(run) > 0)
                if running:
                    break
        else:
            running = len(res) > 0

        return running


@ns.route('/report',
          '/<server>/report',
          endpoint='clients_report')
@ns.doc(
    params={
        'server': 'Which server to collect data from when in multi-agent mode',
    },
)
class ClientsReport(Resource):
    """The :class:`burpui.api.clients.ClientsReport` resource allows you to
    access general reports about your clients.

    This resource is part of the :mod:`burpui.api.clients` module.

    An optional ``GET`` parameter called ``serverName`` is supported when running
    in multi-agent mode.
    """
    parser = ns.parser()
    parser.add_argument('serverName', help='Which server to collect data from when in multi-agent mode')
    stats_fields = ns.model('ClientsStats', {
        'total': fields.Integer(required=True, description='Number of files', default=0),
        'totsize': fields.Integer(required=True, description='Total size occupied by all the backups of this client', default=0),
        'windows': fields.String(required=True, description='Is the client a windows machine', default='unknown'),
    })
    client_fields = ns.model('ClientsReport', {
        'name': fields.String(required=True, description='Client name'),
        'stats': fields.Nested(stats_fields, required=True),
    })
    backup_fields = ns.model('ClientsBackup', {
        'name': fields.String(required=True, description='Client name'),
        'number': fields.Integer(required=True, description='Number of backups on this client', default=0),
    })
    report_fields = ns.model('Report', {
        'backups': fields.Nested(backup_fields, as_list=True, required=True),
        'clients': fields.Nested(client_fields, as_list=True, required=True),
    })

    @api.cache.cached(timeout=1800, key_prefix=cache_key)
    @ns.marshal_with(report_fields, code=200, description='Success')
    @ns.expect(parser)
    @ns.doc(
        responses={
            403: 'Insufficient permissions',
            500: 'Internal failure',
        },
    )
    def get(self, server=None):
        """Returns a global report about all the clients of a given server

        **GET** method provided by the webservice.

        The *JSON* returned is:
        ::

            {
              "backups": [
                {
                  "name": "client1",
                  "number": 15
                },
                {
                  "name": "client2",
                  "number": 1
                }
              ],
              "clients": [
                {
                  "name": "client1",
                  "stats": {
                    "total": 296377,
                    "totsize": 57055793698,
                    "windows": "unknown"
                  }
                },
                {
                  "name": "client2",
                  "stats": {
                    "total": 3117,
                    "totsize": 5345361,
                    "windows": "true"
                  }
                }
              ]
            }


        The output is filtered by the :mod:`burpui.misc.acl` module so that you
        only see stats about the clients you are authorized to.

        :param server: Which server to collect data from when in multi-agent mode
        :type server: str

        :returns: The *JSON* described above
        """

        server = server or self.parser.parse_args()['serverName']
        j = {}
        # Manage ACL
        if (not bui.standalone and bui.acl and
                (not self.is_admin and
                 server not in
                 bui.acl.servers(self.username))):
            self.abort(403, 'Sorry, you don\'t have any rights on this server')
        clients = []
        if bui.acl and not self.is_admin:
            clients = [{'name': x} for x in bui.acl.clients(self.username, server)]
        else:
            try:
                clients = bui.cli.get_all_clients(agent=server)
            except BUIserverException as e:
                self.abort(500, str(e))
        j = bui.cli.get_clients_report(clients, server)
        return j


@ns.route('/stats',
          '/<server>/stats',
          endpoint='clients_stats')
@ns.doc(
    params={
        'server': 'Which server to collect data from when in multi-agent mode',
    },
)
class ClientsStats(Resource):
    """The :class:`burpui.api.clients.ClientsStats` resource allows you to
    access general statistics about your clients.

    This resource is part of the :mod:`burpui.api.clients` module.

    An optional ``GET`` parameter called ``serverName`` is supported when running
    in multi-agent mode.
    """
    parser = ns.parser()
    parser.add_argument('serverName', help='Which server to collect data from when in multi-agent mode')
    client_fields = ns.model('ClientsStatsSingle', {
        'last': fields.DateTime(required=True, dt_format='iso8601', description='Date of last backup'),
        'human': fields.DateTimeHuman(required=True, attribute='last', description='Human readable date of the last backup'),
        'name': fields.String(required=True, description='Client name'),
        'state': fields.String(required=True, description='Current state of the client (idle, backup, etc.)'),
        'phase': fields.String(description='Phase of the current running backup'),
        'percent': fields.Integer(description='Percentage done', default=0),
    })

    @api.cache.cached(timeout=1800, key_prefix=cache_key)
    @ns.marshal_list_with(client_fields, code=200, description='Success')
    @ns.expect(parser)
    @ns.doc(
        responses={
            403: 'Insufficient permissions',
            500: 'Internal failure',
        },
    )
    def get(self, server=None):
        """Returns a list of clients with their states

        **GET** method provided by the webservice.

        The *JSON* returned is:
        ::

            {
              [
                {
                  "last": "2015-05-17 11:40:02",
                  "name": "client1",
                  "state": "idle",
                  "phase": "phase1",
                  "percent": 12,
                },
                {
                  "last": "never",
                  "name": "client2",
                  "state": "idle",
                  "phase": "phase2",
                  "percent": 42,
                }
              ]
            }


        The output is filtered by the :mod:`burpui.misc.acl` module so that you
        only see stats about the clients you are authorized to.

        :param server: Which server to collect data from when in multi-agent mode
        :type server: str

        :returns: The *JSON* described above
        """

        server = server or self.parser.parse_args()['serverName']
        try:
            if (not bui.standalone and
                    bui.acl and
                    (not self.is_admin and
                     server not in
                     bui.acl.servers(self.username))):
                self.abort(403, 'Sorry, you don\'t have any rights on this server')
            j = bui.cli.get_all_clients(agent=server)
            if bui.acl and not self.is_admin:
                j = [x for x in j if x['name'] in bui.acl.clients(self.username, server)]
        except BUIserverException as e:
            self.abort(500, str(e))
        return j


@ns.route('/all',
          '/<server>/all',
          endpoint='clients_all')
@ns.doc(
    params={
        'server': 'Which server to collect data from when in multi-agent mode',
    },
)
class AllClients(Resource):
    """The :class:`burpui.api.clients.AllClients` resource allows you to
    retrieve a list of all clients with their associated server (if any).

    An optional ``GET`` parameter called ``serverName`` is supported when
    running in multi-agent mode.
    """
    parser = ns.parser()
    parser.add_argument('serverName', help='Which server to collect data from when in multi-agent mode')
    client_fields = ns.model('AllClients', {
        'name': fields.String(required=True, description='Client name'),
        'agent': fields.String(required=False, default=None, description='Associated Agent name'),
    })

    @api.cache.cached(timeout=1800, key_prefix=cache_key)
    @ns.marshal_list_with(client_fields, code=200, description='Success')
    @ns.expect(parser)
    @ns.doc(
        responses={
            200: 'Success',
            403: 'Insufficient permissions',
        },
    )
    def get(self, server=None):
        """Returns a list of all clients with their associated Agent if any

        **GET** method provided by the webservice.

        The *JSON* returned is:

        ::

            [
              {
                "name": "client1",
                "agent": "agent1"
              },
              {
                "name": "client2",
                "agent": "agent1"
              },
              {
                "name": "client3",
                "agent": "agent2"
              }
            ]


        The output is filtered by the :mod:`burpui.misc.acl` module so that you
        only see stats about the clients you are authorized to.

        :param server: Which server to collect data from when in multi-agent mode
        :type server: str

        :returns: The *JSON* described above
        """
        ret = []
        args = self.parser.parse_args()
        server = server or args['serverName']

        if (server and bui.acl and not self.is_admin and
                server not in bui.acl.servers(self.username)):
            self.abort(403, "You are not allowed to view this server infos")

        if server:
            clients = bui.cli.get_all_clients(agent=server)
            if bui.acl and not self.is_admin:
                ret = [{'name': x, 'agent': server} for x in bui.acl.clients(self.username, server)]
            else:
                ret = [{'name': x['name'], 'agent': server} for x in clients]
            return ret

        if bui.standalone:
            if bui.acl and not self.is_admin:
                ret = [{'name': x} for x in bui.acl.clients(self.username)]
            else:
                ret = [{'name': x['name']} for x in bui.cli.get_all_clients()]
        else:
            grants = {}
            if bui.acl and not self.is_admin:
                for serv in bui.acl.servers(self.username):
                    grants[serv] = bui.acl.clients(self.username, serv)
            else:
                for serv in bui.cli.servers:
                    grants[serv] = 'all'
            for (serv, clients) in iteritems(grants):
                if not isinstance(clients, list):
                    clients = [x['name'] for x in bui.cli.get_all_clients(agent=serv)]
                ret += [{'name': x, 'agent': serv} for x in clients]

        return ret



# -*- coding: utf8 -*-
import re
import codecs

from .interface import BUIhandler, BUIuser, BUIloader
from ...utils import NOTIF_ERROR, NOTIF_OK, NOTIF_WARN
from werkzeug.security import check_password_hash, generate_password_hash


class BasicLoader(BUIloader):
    """The :class:`burpui.misc.auth.basic.BasicLoader` class loads the *Basic*
    users.
    """
    section = name = 'BASIC'

    def __init__(self, app=None, handler=None):
        """:func:`burpui.misc.auth.basic.BasicLoader.__init__` loads users from
        the configuration file.

        :param app: Instance of the app we are running in
        :type app: :class:`burpui.server.BUIServer`

        :param handler: Instance of handler
        :type handler: :class:`burpui.misc.auth.interface.BUIhandler`
        """
        self.app = app
        self.conf = self.app.conf
        self.conf_id = None
        self.users = {}
        self.handler = handler
        self.handler.name = self.name
        self.handler.add_user = self.add_user
        self.handler.del_user = self.del_user
        self.handler.change_password = self.change_password
        self.load_users(True)

    def load_users(self, force=False):
        if not force and self.conf_id:
            if not self.conf.changed(self.conf_id):
                return False

        self.users = {
            'admin': {'pwd': generate_password_hash('admin'), 'salted': True}
        }

        if self.section in self.conf.options:
            # check passwords are salted
            salted = False
            changed = False
            # This is not necessary for now. Maybe will use this some day
            # TODO: clean this?
            # if len(self.conf.options.comments[self.section]) > 0:
            #     if re.match(
            #             r'^\s*#+\s*@salted@',
            #             self.conf.options.comments[self.section][-1]):
            #         salted = True
            # allow mixed logins (plain and hashed)
            # TODO: defaults to False for 0.4.0
            mixed = self.conf.safe_get(
                'mixed',
                cast='boolean',
                section=self.section,
                defaults=True
            )
            self.users = {}
            for opt in self.conf.options.get(self.section).keys():
                salt = True
                if opt == 'priority':
                    # Maybe the handler argument is None, maybe the 'priority'
                    # option is missing. We don't care.
                    try:
                        self.handler.priority = self.conf.safe_get(
                            opt,
                            section=self.section
                        ) or 0
                    except:
                        pass
                    continue  # pragma: no cover
                # list of reserved options
                if opt in ['mixed']:
                    continue
                pwd = self.conf.safe_get(opt, section=self.section)
                if not salted or mixed:
                    if not re.match(r'^pbkdf2:.+\$.+\$.+', pwd):
                        if mixed:
                            salt = False
                        else:
                            # mixed not allowed so we convert plain passwords
                            pwd = generate_password_hash(pwd)
                            self.conf.options[self.section][opt] = pwd
                            changed = True
                self.users[opt] = {'pwd': pwd, 'salted': salt}
                self.logger.info('Loading user: {} ({})'.format(
                    opt,
                    'hashed' if salt else 'plain')
                )

            if changed:
                self.conf.options.write()
            # if not salted:
            #     self.conf.options.comments[self.section].append(
            #         '# Please DO NOT touch the following line'
            #     )
            #     self.conf.options.comments[self.section].append('# @salted@')
            #     self.conf.options.write()
            self.conf_id = self.conf.id
        return True

    def fetch(self, uid=None):
        """:func:`burpui.misc.auth.basic.BasicLoader.fetch` searches for a user
        in the configuration.

        :param uid: User to search for
        :type uid: str

        :returns: The given UID if the user exists or None
        """
        self.load_users()
        if uid in self.users:
            return uid

        return None

    def check(self, uid=None, passwd=None):
        """:func:`burpui.misc.auth.basic.BasicLoader.check` verifies if the
        given password matches the given user settings.

        :param uid: User to authenticate
        :type uid: str

        :param passwd: Password
        :type passwd: str

        :returns: True if there is a match, otherwise False
        """
        self.load_users()
        if uid in self.users:
            if self.users[uid]['salted']:
                return check_password_hash(self.users[uid]['pwd'], passwd)
            else:
                return self.users[uid]['pwd'] == passwd
        return False

    def _setup_users(self):
        """Setup user management"""
        if self.section not in self.conf.options:
            # look for the section in the comments
            conffile = self.conf.options.filename
            # last = ''
            ori = []
            with codecs.open(conffile, 'r', 'utf-8') as config:
                ori = [x.rstrip('\n') for x in config.readlines()]
            if ori:
                with codecs.open(conffile, 'w', 'utf-8') as config:
                    found = False
                    for line in ori:
                        if re.match(r'^\s*#+\s*\[{}\]'.format(self.section),
                                    line):
                            # Not useful for now
                            # TODO: clean this?
                            # if not last or \
                            #         not re.match(r'^\s*#+\s*@salted@', last):
                            #     config.write(
                            #         '# Please DO NOT touch the following line\n'
                            #     )
                            #     config.write('# @salted@\n')

                            config.write('[{}]\n'.format(self.section))
                            found = True
                        else:
                            config.write('{}\n'.format(line))
                            # last = line

                    if not found:
                        # config.write(
                        #     '# Please DO NOT touch the following line\n'
                        # )
                        # config.write('# @salted@\n')
                        config.write('[{}]\n'.format(self.section))

            self.conf.options.reload()

    def add_user(self, user, passwd):
        """Add a user"""
        self._setup_users()
        if user in self.users:
            message = "user '{}' already exists".format(user)
            self.logger.warning(message)
            return False, message, NOTIF_WARN
        pwd = generate_password_hash(passwd)
        self.conf.options[self.section][user] = pwd
        self.conf.options.write()
        self.load_users(True)
        message = "user '{}' successfully added".format(user)
        return True, message, NOTIF_OK

    def del_user(self, user):
        """Delete a user"""
        self._setup_users()
        self.load_users(True)
        if user not in self.users:
            message = "user '{}' does not exist".format(user)
            self.logger.error(message)
            return False, message, NOTIF_ERROR
        if user == 'admin' and len(self.users.keys()) == 1:
            message = 'trying to delete the admin account!'
            self.logger.warning(message)
            return False, message, NOTIF_WARN
        del self.conf.options[self.section][user]
        del self.users[user]
        self.conf.options.write()
        self.load_users(True)
        message = "user '{}' successfully removed".format(user)
        return True, message, NOTIF_OK

    def change_password(self, user, passwd):
        """Change a user password"""
        self._setup_users()
        self.load_users(True)
        if user not in self.users:
            message = "user '{}' does not exist".format(user)
            self.logger.error(message)
            return False, message, NOTIF_ERROR
        current = self.users[user]
        if current['salted']:
            func = check_password_hash
        else:
            def func(x, y):
                return x == y
        curr = current['pwd']
        if func(curr, passwd):
            message = 'password is the same'
            self.logger.warning(message)
            return False, message, NOTIF_WARN
        pwd = generate_password_hash(passwd)
        self.conf.options[self.section][user] = pwd
        self.conf.options.write()
        self.load_users(True)
        message = "user '{}' successfully updated".format(user)
        return True, message, NOTIF_OK


class UserHandler(BUIhandler):
    """See :class:`burpui.misc.auth.interface.BUIhandler`"""
    def __init__(self, app=None):
        """See :func:`burpui.misc.auth.interface.BUIhandler.__init__`"""
        self.basic = BasicLoader(app, self)
        self.change = False
        self.users = {}

    def user(self, name=None):
        """See :func:`burpui.misc.auth.interface.BUIhandler.user`"""
        if name not in self.users:
            self.change = self.basic.load_users()
            self.users[name] = BasicUser(self.basic, name)
        return self.users[name]

    @property
    def changed(self):
        ret = self.change or self.basic.load_users()
        self.change = False
        return ret

    @property
    def loader(self):
        return self.basic


class BasicUser(BUIuser):
    """See :class:`burpui.misc.auth.interface.BUIuser`"""
    def __init__(self, basic=None, name=None):
        self.active = False
        self.authenticated = False
        self.basic = basic
        self.name = name
        self.id = None
        self.backend = self.basic.name

        res = self.basic.fetch(self.name)

        if res:
            self.id = res
            self.active = True

    def login(self, passwd=None):
        """See :func:`burpui.misc.auth.interface.BUIuser.login`"""
        self.authenticated = self.basic.check(self.name, passwd)
        return self.authenticated

    def get_id(self):
        return self.id



# -*- coding: utf8 -*-
"""
.. module:: burpui.config
    :platform: Unix
    :synopsis: Burp-UI config module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>

"""
import os
import re
import json
import codecs
import shutil
import datetime
import logging
import configobj
import validate

from . import __version__, __release__


class BUIConfig(dict):
    """Custom config parser"""
    logger = logging.getLogger('burp-ui')
    delta = datetime.timedelta(seconds=30)
    last = datetime.datetime.now() - delta
    mtime = 0

    def __init__(self, config=None, explain=False, defaults=None):
        """Wrapper around the ConfigObj class

        :param config: Configuration to parse
        :type config: str, list or File

        :param explain: Whether to explain the parsing errors or not
        :type explain: bool

        :param defaults: Default options
        :type defaults: dict
        """
        if config:
            self.parse(config, explain, defaults)

    def parse(self, config, explain=False, defaults=None):
        """Parse the conf

        :param config: Configuration to parse
        :type config: str, list or File

        :param explain: Whether to explain the parsing errors or not
        :type explain: bool

        :param defaults: Default options
        :type defaults: dict
        """
        self.conf = {}
        self.conffile = config
        self.section = None
        self.defaults = defaults
        self.validator = validate.Validator()
        try:
            self.conf = configobj.ConfigObj(config, encoding='utf-8')
            self.last = datetime.datetime.now()
            self.mtime = os.path.getmtime(self.conffile)
        except configobj.ConfigObjError as exp:
            # We were unable to parse the config, maybe we need to
            # convert/update it
            self.logger.warning(
                'Unable to parse the configuration... Trying to convert it'
            )
            # if conversion is successful, give it another try
            if self._convert(config):
                # This time, if it fails, the exception will be forwarded
                try:
                    self.conf = configobj.ConfigObj(config)
                except configobj.ConfigObjError as exp2:
                    if explain:
                        self._explain(exp2)
                    else:
                        raise exp2
            else:
                self.logger.critical('Unable to convert configuration')
                if explain:
                    self._explain(exp)
                else:
                    raise exp

    @property
    def options(self):
        """ConfigObj object"""
        if (datetime.datetime.now() - self.last) > self.delta:
            self._refresh()
        return self.conf

    @property
    def id(self):
        """Conf ID to detect changes"""
        return self.mtime

    @staticmethod
    def string_lower_list(value):
        if not value:
            raise validate.VdtMissingValue('No value for this option')
        if not isinstance(value, list):
            return [str(value).lower()]
        return [str(x).lower() for x in value]

    def boolean_or_string(self, value):
        if not value:
            raise validate.VdtMissingValue('No value for this option')
        try:
            caster = self.validator.functions.get('boolean')
            return caster(value)
        except validate.VdtTypeError:
            return value

    def changed(self, id):
        """Check if the conf has changed"""
        if (datetime.datetime.now() - self.last) > self.delta:
            self._refresh()
        return id != self.mtime

    def _refresh(self):
        """Refresh conf"""
        self.last = datetime.datetime.now()
        mtime = os.path.getmtime(self.conffile)
        if mtime != self.mtime:
            self.logger.debug('Configuration changed')
            self.mtime = mtime
            self.conf.reload()

    def update_defaults(self, new_defaults):
        """Add new defaults"""
        self.defaults.update(new_defaults)

    def default_section(self, section):
        """Set the default section"""
        self.section = section

    def _convert(self, config):
        """Convert an old config to a new one"""
        sav = '{}.back'.format(config)
        current_section = None

        if os.path.exists(sav):
            self.logger.error(
                'Looks like the configuration file has already been converted'
            )
            return False

        try:
            shutil.copy(config, sav)
        except IOError as exp:
            self.logger.error(str(exp))
            return False

        try:
            with codecs.open(sav, 'r', 'utf-8') as ori:
                with codecs.open(config, 'w', 'utf-8') as new:
                    # We add some headers
                    new.write('# Auto-generated file from a previous version\n')
                    new.write('# @version@ - {}\n'.format(__version__))
                    new.write('# @release@ - {}\n'.format(__release__))
                    for line in ori.readlines():
                        line = line.rstrip('\n')
                        search = re.search(r'^\s*(#?)\s*\[([^\]]+)\]\s*', line)
                        if search:
                            if not search.group(1):
                                current_section = search.group(2)
                        # if we find old style config lines, we convert them
                        elif re.match(r'^\s*\S+\s*:\s*.+$', line) and \
                                re.match(r'^\s*[^\[]', line):
                            key, val = re.split(r'\s*:\s*', line, 1)
                            # We support *objects* but we need to serialize them
                            try:
                                jsn = json.loads(val)
                                # special case, we re-format the admin value
                                if current_section == 'BASIC:ACL' and \
                                        key == 'admin' and \
                                        isinstance(jsn, list):
                                    val = ','.join(jsn)
                                elif isinstance(jsn, list) or \
                                        isinstance(jsn, dict):
                                    val = "'{}'".format(json.dumps(jsn))
                            except ValueError:
                                pass
                            line = '{} = {}'.format(key, val)

                        new.write('{}\n'.format(line))

        except IOError as exp:
            self.logger.error(str(exp))
            return False
        return True

    @staticmethod
    def _explain(exception):
        """Explain parsing errors

        :param exception: Exception object
        :type exception: :class:`configobj.ConfigObjError`
        """
        message = u'\n'
        for error in exception.errors:
            message += error.message + '\n'

        raise configobj.ConfigObjError(message.rstrip('\n'))

    def safe_get(
            self,
            key,
            cast='pass',
            section=None,
            defaults=None):
        """Safely return the asked option

        :param key: Key name
        :type key: str

        :param cast: How to cast the option
        :type cast: str

        :param section: Section name
        :type section: str

        :param defaults: Default options
        :type defaults: dict, mixed

        :returns: The value of the asked option
        """
        # The configobj validator is sooo broken. We need to workaround it...
        default_by_type = {
            'integer': 0,
            'boolean': False,
        }
        section = section or self.section
        # if the defaults argument is not a list, assume it's a single value
        if defaults and not isinstance(defaults, dict):
            defaults = {section: {key: defaults}}
        defaults = defaults or self.defaults
        if section not in self.conf:
            self.logger.warning("No '{}' section found".format(section))
            if defaults:
                return defaults.get(section, {}).get(key)
            return None

        val = self.conf.get(section).get(key)
        default = default_by_type.get(cast)
        if defaults and section in defaults and \
                key in defaults.get(section):
            default = defaults.get(section, {}).get(key)
        try:
            caster = self.validator.functions.get(cast)
            if not caster:
                try:
                    caster = getattr(self, cast)
                except AttributeError:
                    self.logger.error(
                        "'{}': no such validator".format(cast)
                    )
                    return val
            ret = caster(val)
            # special case for boolean and integer, etc.
            if ret is None:
                ret = default
            self.logger.debug(
                '[{}]:{} - found: {}, default: {} -> {}'.format(
                    section,
                    key,
                    val,
                    default,
                    ret
                )
            )
        except validate.ValidateError as exp:
            ret = default
            self.logger.warning(
                '{}\n[{}]:{} - found: {}, default: {} -> {}'.format(
                    str(exp),
                    section,
                    key,
                    val,
                    default,
                    ret
                )
            )
        return ret


config = BUIConfig()



# -*- coding: utf8 -*-
"""
.. module:: burpui.ext.sql
    :platform: Unix
    :synopsis: Burp-UI external SQL module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>

"""
from flask_sqlalchemy import SQLAlchemy

db = SQLAlchemy()



# -*- coding: utf8 -*-
"""
.. module:: burpui.misc.parser.burp2
    :platform: Unix
    :synopsis: Burp-UI configuration file parser for Burp2.
.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>
"""
from .burp1 import Parser as Burp1


# inherit Burp1 parser so we can just override available options
class Parser(Burp1):
    """Extends :class:`burpui.misc.parser.burp1.Parser`"""
    pver = 2

    multi_srv = Burp1.multi_srv + [
        u'label',
    ]
    string_srv = Burp1.string_srv + [
        u'manual_delete',
    ]
    boolean_add = [
        u'acl',
        u'xattr',
        u'server_can_override_includes',
        u'glob_after_script_pre',
    ]
    boolean_add_cli = [
        u'enabled',
    ]
    boolean_srv = Burp1.boolean_srv + boolean_add
    boolean_cli = Burp1.boolean_cli + boolean_add + boolean_add_cli
    multi_cli = Burp1.multi_cli + [
        u'label',
    ]
    integer_cli = Burp1.integer_cli + [
        u'randomise',
    ]
    fields_cli = Burp1.fields_cli + boolean_add + boolean_add_cli + [
        u'randomise',
        u'manual_delete',
        u'label',
    ]
    placeholders = Burp1.placeholders
    placeholders.update({
        u'acl': u"0|1",
        u'xattr': u"0|1",
        u'randomise': u"max secs",
        u'manual_delete': u"path",
        u'label': u"some informations",
        u'server_can_override_includes': u"0|1",
        u'glob_after_script_pre': u"0|1",
        u'enabled': u"0|1",
    })
    defaults = Burp1.defaults
    defaults.update({
        u'acl': True,
        u'xattr': True,
        u'server_can_override_includes': True,
        u'glob_after_script_pre': True,
        u'randomise': 0,
        u'manual_delete': u'',
        u'label': u'',
        u'enabled': True,
    })
    doc = Burp1.doc
    doc.update({
        u'acl': u"If acl support is compiled into burp, this allows you to"
                " decide whether or not to backup acls at runtime. The"
                " default is '1'.",
        u'xattr': u"If xattr support is compiled into burp, this allows you"
                  " to decide whether or not to backup xattrs at runtime."
                  " The default is '1'.",
        u'randomise': u"When running a timed backup, sleep for a random"
                      " number of seconds (between 0 and the number given)"
                      " before contacting the server. Alternatively, this"
                      " can be specified by the '-q' command line option.",
        u'manual_delete': u"This can be overridden by the clientconfdir"
                          " configuration files in clientconfdir on the"
                          " server. When the server needs to delete old"
                          " backups, or rubble left over from generating"
                          " reverse patches with librsync=1, it will"
                          " normally delete them in place. If you use the"
                          " 'manual_delete' option, the files will be moved"
                          " to the path specified for deletion at a later"
                          " point. You will then need to configure a cron"
                          " job, or similar, to delete the files yourself."
                          " Do not specify a path that is not on the same"
                          " filesystem as the client storage directory.",
        u'label': u"You can have multiple labels, and they can be overridden"
                  " in the client configuration files in clientconfdir on"
                  " the server. They will appear as an array of strings in"
                  " the server status monitor JSON output. The idea is to"
                  " provide a mechanism for arbitrary values to be passed"
                  " to clients of the server status monitor.",
        u'server_can_override_includes': u"To prevent the server from being"
                                         " able to override your local"
                                         " include/exclude list, set this"
                                         " to 0. The default is 1.",
        u'glob_after_script_pre': u"Set this to 0 if you do not want"
                                  " include_glob settings to be evaluated"
                                  " after the pre script is run. The default"
                                  " is 1.",
        u'enabled': u"Set this to 0 if you want to disable all clients. The"
                    " default is 1. This option can be overridden per-client"
                    " in the client configuration files in clientconfdir on"
                    " the server.",
    })



# -*- coding: utf8 -*-
"""
.. module:: burpui.misc.parser.interface
    :platform: Unix
    :synopsis: Burp-UI parser interface.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>

"""
from abc import ABCMeta, abstractmethod
from six import with_metaclass

import logging


class BUIparser(with_metaclass(ABCMeta, object)):
    """:class:`burpui.misc.parser.interface.BUIparser` defines a generic
    interface for ``burp`` configuration files parser.
    """

    logger = logging.getLogger('burp-ui')

    @abstractmethod
    def read_server_conf(self, conf=None):
        """:func:`burpui.misc.parser.interface.BUIparser.read_server_conf` is
        called by :func:`burpui.misc.backend.interface.BUIbackend.read_conf_srv`
        in order to parse the burp-server configuration file.

        :param conf: Complementary configuration file (for instance, file
                     inclusions)
        :type conf: str

        :returns: Dict of options

        Example::

            {
                "boolean": [
                    {
                        "name": "hardlinked_archive",
                        "value": false
                    },
                    {
                        "name": "syslog",
                        "value": true
                    },
                ],
                "clients": [
                    {
                        "name": "client1",
                        "value": "/etc/burp/clientconfdir/client1"
                    },
                    {
                        "name": "client2",
                        "value": "/etc/burp/clientconfdir/client2"
                    },
                ],
                "common": [
                    {
                        "name": "mode",
                        "value": "server"
                    },
                    {
                        "name": "directory",
                        "value": "/srv/burp"
                    },
                ],
                "includes": [],
                "includes_ext": [],
                "integer": [
                    {
                        "name": "port",
                        "value": 4971
                    },
                    {
                        "name": "status_port",
                        "value": 4972
                    },
                    {
                        "name": "max_children",
                        "value": 5
                    },
                    {
                        "name": "max_status_children",
                        "value": 5
                    }
                ],
                "multi": [
                    {
                        "name": "keep",
                        "value": [
                            "7",
                            "4",
                            "4"
                        ]
                    },
                    {
                        "name": "timer_arg",
                        "value": [
                            "12h",
                            "Mon,Tue,Thu,Fri,17,18,19,20,21,22,23",
                            "Wed,Sat,Sun,06,07,08,09,10,11,12,13,14,15,16,17,18,19,20,21,22,23"
                        ]
                    },
                ],
            }
        """
        raise NotImplementedError(
            "Sorry, the current Parser does not implement this method!"
        )  # pragma: no cover

    @abstractmethod
    def store_client_conf(self, data, client=None, conf=None):
        """:func:`burpui.misc.parser.interface.BUIparser.store_client_conf` is
        used by :func:`burpui.misc.backend.BUIbackend.store_conf_cli`.

        It works the same way as
        :func:`burpui.misc.parser.interface.BUIparser.store_conf`
        with an extra parameter:

        :param client: Name of the client for which to apply this config
        :type client: str
        """
        raise NotImplementedError(
            "Sorry, the current Parser does not implement this method!"
        )  # pragma: no cover

    @abstractmethod
    def store_conf(self, data, conf=None, client=None, mode='srv'):
        """:func:`burpui.misc.parser.interface.BUIparser.store_conf` is used to
        store the configuration from the web-ui into the actual configuration
        files.
        It is used by :func:`burpui.misc.backend.BUIbackend.store_conf_srv`.

        :param data: Data sent by the web-form
        :type data: dict

        :param conf: Force the file path (for file inclusions for instance)
        :type conf: str

        :param client: Client name
        :type client: str

        :param mode: We actually use the same method for clients and server
                     files
        :type mode: str

        :returns: A list of notifications to return to the UI (success or
                  failure)

        Example::

            [[0, "Success"]]
        """
        raise NotImplementedError(
            "Sorry, the current Parser does not implement this method!"
        )  # pragma: no cover

    @abstractmethod
    def path_expander(self, pattern=None, source=None, client=None):
        """:func:`burpui.misc.parser.interface.BUIparser.path_expander` is used
        to expand path of file inclusions glob the user can set in the setting
        panel.

        :param pattern: The glob/path to expand
        :type pattern: str

        :param source: What file we are working in
        :type source: str

        :param client: The client name when working on client files
        :type client: str

        :returns: A list of files or an empty list
        """
        raise NotImplementedError(
            "Sorry, the current Parser does not implement this method!"
        )  # pragma: no cover

    @abstractmethod
    def list_clients(self):
        """:func:`burpui.misc.parser.interface.BUIparser.list_clients` is used
        to retrieve a list of clients with their configuration file.

        :returns: A list of clients with their configuration file
        """
        raise NotImplementedError(
            "Sorry, the current Parser does not implement this method!"
        )  # pragma: no cover

    @abstractmethod
    def is_client_revoked(self, client=None):
        """:func:`burpui.misc.parser.interface.BUIparser.is_client_revoked` is
        used to check if a given client has it's certificate revoked or not.

        :param client: The name of the client to check
        :type client: str

        :returns: True or False
        """
        raise NotImplementedError(
            "Sorry, the current Parser does not implement this method!"
        )  # pragma: no cover

    @abstractmethod
    def remove_client(self, client=None, delcert=False, revoke=False):
        """:func:`burpui.misc.parser.interface.BUIparser.remove_client` is used
        to delete a client from burp's configuration.

        :param client: The name of the client to remove
        :type client: str

        :param delcert: Whether to delete the associated certificate
        :type delcert: bool

        :param revoke: Whether to revoke the associated certificate
        :type revoke: bool

        :returns: A list of notifications to return to the UI (success or
                  failure)
        """
        raise NotImplementedError(
            "Sorry, the current Parser does not implement this method!"
        )  # pragma: no cover

    @abstractmethod
    def read_client_conf(self, client=None, conf=None):
        """:func:`burpui.misc.parser.interface.BUIparser.read_client_conf` is
        called by :func:`burpui.misc.backend.interface.BUIbackend.read_conf_cli`
        in order to parse the burp-clients configuration files.

        It works the same way as
        :func:`burpui.misc.parser.interface.BUIparser.read_server_conf`
        """
        raise NotImplementedError(
            "Sorry, the current Parser does not implement this method!"
        )  # pragma: no cover

    @abstractmethod
    def cancel_restore(self, name=None):
        """:func:`burpui.misc.parser.interface.BUIparser.cancel_restore` called
        by
        :func:`burpui.misc.backend.interface.BUIbackend.cancel_server_restore`
        in order to cancel a server-initiated restoration.

        :param name: Client name
        :type name: str

        :returns: A list of notifications to return to the UI (success or
                  failure)
        """
        raise NotImplementedError(
            "Sorry, the current Parser does not implement this method!"
        )  # pragma: no cover

    @abstractmethod
    def read_restore(self, name=None):
        """:func:`burpui.misc.parser.interface.BUIparser.read_restore` called
        by :func:`burpui.misc.backend.interface.BUIbackend.is_server_restore`
        in order to read a server-initiated restoration file.

        :param name: Client name
        :type name: str

        :returns: A dict describing the content of the file
        """
        raise NotImplementedError(
            "Sorry, the current Parser does not implement this method!"
        )  # pragma: no cover

    @abstractmethod
    def server_initiated_restoration(
            self,
            name=None,
            backup=None,
            files=None,
            strip=None,
            force=None,
            prefix=None,
            restoreto=None):
        """
        :func:`burpui.misc.parser.interface.BUIparser.server_initiated_restoration`
        called by :func:`burpui.misc.backend.interface.BUIbackend.server_restore`
        in order to create server-initiated restoration file.

        :param name: Client name
        :type name: str

        :param backup: Backup number
        :type backup: int

        :param files: List of files to restore
        :type files: str

        :param strip: Number of leading path to strip
        :type strip: int

        :param force: Whether to force overriding files or not
        :type force: bool

        :param prefix: Where to restore files
        :type prefix: str

        :param agent: What server to ask (only in multi-agent mode)
        :type agent: str

        :returns: A list of notifications to return to the UI (success or
                  failure)
        """
        raise NotImplementedError(
            "Sorry, the current Parser does not implement this method!"
        )  # pragma: no cover

    @abstractmethod
    def cancel_backup(self, name=None):
        """:func:`burpui.misc.parser.interface.BUIparser.cancel_backup` called
        by :func:`burpui.misc.backend.interface.BUIbackend.cancel_server_backup`
        in order to cancel a server-initiated backup.

        :param name: Client name
        :type name: str

        :returns: A list of notifications to return to the UI (success or
                  failure)
        """
        raise NotImplementedError(
            "Sorry, the current Parser does not implement this method!"
        )  # pragma: no cover

    @abstractmethod
    def read_backup(self, name=None):
        """:func:`burpui.misc.parser.interface.BUIparser.read_backup` called
        by :func:`burpui.misc.backend.interface.BUIbackend.is_server_backup`
        in order to test the existence of a server-initiated backup file.

        :param name: Client name
        :type name: str

        :returns: A True if the file is found, else False.
        """
        raise NotImplementedError(
            "Sorry, the current Parser does not implement this method!"
        )  # pragma: no cover

    @abstractmethod
    def server_initiated_backup(self, name=None):
        """:func:`burpui.misc.parser.interface.BUIparser.server_initiated_backup`
        called by :func:`burpui.misc.backend.interface.BUIbackend.server_backup`
        in order to create a server-initiated backup file.

        :param name: Client name
        :type name: str

        :returns: A list of notifications to return to the UI (success or
                  failure)
        """
        raise NotImplementedError(
            "Sorry, the current Parser does not implement this method!"
        )  # pragma: no cover

    @abstractmethod
    def param(self, name, obj='server_conf', client=None):
        """:func:`burpui.misc.parser.interface.BUIparser.param`
        lookup for a given param in the conf.

        :param name: Param name
        :type name: str

        :param obj: Object to look param for
        :type obj: str

        :param client: Search for a given client param
        :type client: str

        :returns: The asked param
        """
        raise NotImplementedError(
            "Sorry, the current Parser does not implement this method!"
        )  # pragma: no cover



# -*- coding: utf8 -*-
"""
.. module:: burpui.api.custom.resource
    :platform: Unix
    :synopsis: Burp-UI api custom resource module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>

"""
import logging
import inspect
import json

from flask_restplus import Resource as ResourcePlus
from flask_restplus.errors import abort
from flask_login import current_user
from flask import current_app as bui
from ..._compat import PY3

if PY3:
    basestring = str


class Resource(ResourcePlus):
    """Subclass default Resource to manage ACL"""
    logger = logging.getLogger('burp-ui')

    def __init__(self, api=None, *args, **kwargs):
        """Constructor"""
        self.username = current_user.get_id()
        # if there is no ACL, assume everybody is admin
        self.is_admin = True
        if bui.acl:
            self.is_admin = bui.acl.is_admin(self.username)
        self.cache = api.cache
        ResourcePlus.__init__(self, api, *args, **kwargs)

    def abort(self, code=500, message=None, **kwargs):
        """
        Properly abort the current request

        See: :func:`~flask_restplus.errors.abort`
        """
        if message and not isinstance(message, basestring):
            try:
                message = json.dumps(message)  # pragma: no cover
            except:
                message = None
        # Add extra logs when raising abort exception
        (
            frm,
            filename,
            line_no,
            func,
            source_code,
            source_index
        ) = inspect.stack()[1]
        mod = inspect.getmodule(frm)
        self.logger.debug('Abort in {}:{}'.format(filename, line_no))
        self.logger.warning(
            '[{}] {}: {}{}'.format(
                mod.__name__,
                code,
                message,
                ' - {}'.format(kwargs) if kwargs else ''
            )
        )
        # This raises a Flask Exception
        abort(code, message, **kwargs)



# -*- coding: utf8 -*-
"""
.. module:: burpui.ext.cache
    :platform: Unix
    :synopsis: Burp-UI external cache module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>

"""
import warnings

from flask_cache import Cache as CacheOrig
from flask.exthook import ExtDeprecationWarning


class Cache(CacheOrig):
    def init_app(self, app, config=None):
        # suppress warnings since upstream did not fix it yet
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', ExtDeprecationWarning)
            super(Cache, self).init_app(app, config)

cache = Cache(config={'CACHE_TYPE': 'null', 'CACHE_NO_NULL_WARNING': True})



# -*- coding: utf8 -*-
"""
.. module:: burpui._json
    :platform: Unix
    :synopsis: Burp-UI json compatibility module.

.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>

"""
import ujson
from six import viewkeys

__implements__ = ['dumps', 'loads']
ori_dumps = None
ori_loads = None

IMPLEMENTED_DUMPS_KWARGS = [
    'ensure_ascii',
    'double_precision',
    'encode_html_chars',
    'sort_keys',
]
IMPLEMENTED_LOADS_KWARGS = [
    'precise_float',
]


# ujson does not implement all the features of the original json parser
# the trick here is to catch such an exception to fallback to the original one
def dumps(*args, **kwargs):
    keys = []
    if kwargs:
        keys = viewkeys(kwargs)
    for key in keys:
        if key not in IMPLEMENTED_DUMPS_KWARGS:
            return ori_dumps(*args, **kwargs)
    try:
        return ujson.dumps(*args, **kwargs)
    except:
        return ori_dumps(*args, **kwargs)


def loads(*args, **kwargs):
    keys = []
    if kwargs:
        keys = viewkeys(kwargs)
    for key in keys:
        if key not in IMPLEMENTED_LOADS_KWARGS:
            return ori_loads(*args, **kwargs)
    try:
        return ujson.loads(*args, **kwargs)
    except:
        return ori_loads(*args, **kwargs)



# -*- coding: utf8 -*-
"""
.. module:: burpui.misc.parser.doc
    :platform: Unix
    :synopsis: Burp-UI parser documentation for Burp1.
.. moduleauthor:: Ziirish <hi+burpui@ziirish.me>
"""
from .interface import BUIparser


class Doc(BUIparser):
    """:class:`burpui.misc.parser.doc.Doc` provides a consistent interface
    to parse burp configuration files.

    It implements :class:`burpui.misc.parser.interface.BUIparser`.
    """
    defaults = {
        u'address': u'',  # IP
        u'atime': False,  # bool
        u'autoupgrade_dir': u'',  # dir
        u'ca_burp_ca': u'',  # file
        u'ca_conf': u'',  # file
        u'ca_name': u'',
        u'ca_server_name': u'',
        u'client_can_delete': True,  # bool
        u'client_can_diff': True,  # bool
        u'client_can_force_backup': True,  # bool
        u'client_can_list': True,  # bool
        u'client_can_restore': True,  # bool
        u'client_can_verify': True,  # bool
        u'client_lockdir': u'',
        u'clientconfdir': u'',  # dir
        u'compression': u'gzip9',
        u'cross_all_filesystems': False,  # bool
        u'cross_filesystem': u'',
        u'daemon': True,  # bool
        u'dedup_group': u'',
        u'directory_tree': True,  # bool
        u'directory': u'',
        u'exclude_comp': u'',  # multi
        u'exclude_ext': u'',  # multi
        u'exclude_fs': u'',  # multi
        u'exclude_regex': u'',  # multi
        u'exclude': u'',  # multi
        u'fork': True,  # bool
        u'group': u'',
        u'hard_quota': u'',
        u'hardlinked_archive': False,  # bool
        u'include_ext': u'',  # multi
        u'include_glob': u'',  # multi
        u'include_regex': u'',  # multi
        u'include': u'',  # multi
        u'keep': [7, 6, 4],  # multi  # int
        u'librsync': True,  # bool
        u'max_children': 5,  # int
        u'max_file_size': u'',
        u'max_hardlinks': 10000,  # int
        u'max_status_children': 5,  # int
        u'max_storage_subdirs': 30000,  # int
        u'min_file_size': u'',
        u'mode': u'',
        u'monitor_browse_cache': False,  # bool
        u'network_timeout': 7200,  # int
        u'nobackup': u'',
        u'notify_failure_arg': u'',  # multi
        u'notify_failure_script': u'',  # file
        u'notify_success_arg': u'',  # multi
        u'notify_success_changes_only': False,  # bool
        u'notify_success_script': u'',  # file
        u'notify_success_warnings_only': False,  # bool
        u'password': u'',
        u'password_check': True,  # bool
        u'path_length_warn': True,  # bool
        u'pidfile': u'',
        u'port': 4971,  # int
        u'protocol': False,  # int
        u'ratelimit': False,  # int
        u'read_all_blockdevs': False,  # bool
        u'read_all_fifos': False,  # bool
        u'read_blockdev': u'',
        u'read_fifo': u'',
        u'restore_client': u'',  # multi
        u'scan_problem_raises_error': False,  # bool
        u'server_script_arg': u'',  # multi
        u'server_script_notify': False,  # bool
        u'server_script_post_arg': u'',  # multi
        u'server_script_post_notify': False,  # bool
        u'server_script_post_run_on_fail': False,  # bool
        u'server_script_post': u'',  # file
        u'server_script_pre_arg': u'',  # multi
        u'server_script_pre_notify': False,  # bool
        u'server_script_pre': u'',  # file
        u'server_script': u'',  # file
        u'soft_quota': u'',
        u'split_vss': False,  # bool
        u'ssl_cert_ca': u'',  # file
        u'ssl_cert': u'',  # file
        u'ssl_ciphers': u'',
        u'ssl_compression': u'zlib5',
        u'ssl_dhfile': u'',  # file
        u'ssl_key_password': u'',
        u'ssl_key': u'',  # file
        u'ssl_peer_cn': u'',
        u'status_address': u'',  # 127.0.0.1 / ::1
        u'status_port': 4972,  # int
        u'stdout': True,  # bool
        u'strip_vss': False,  # bool
        u'syslog': False,  # bool
        u'timer_arg': u'',  # multi
        u'timer_script': u'',  # file
        u'timestamp_format': u'',
        u'umask': u'0022',  # mode
        u'user': u'',
        u'version_warn': True,  # bool
        u'vss_drives': u'',
        u'working_dir_recovery_method': u'',
        u'server_can_restore': False,
    }
    placeholders = {
        u'.': u"path or glob",
        u'atime': u"0|1",
        u'autoupgrade_dir': u"path",
        u'ca_burp_ca': u"path",
        u'ca_conf': u"path",
        u'ca_name': u"name",
        u'ca_server_name': u"name",
        u'client_can_delete': u"0|1",
        u'client_can_force_backup': u"0|1",
        u'client_can_list': u"0|1",
        u'client_can_restore': u"0|1",
        u'client_can_verify': u"0|1",
        u'client_lockdir': u"path",
        u'clientconfdir': u"path",
        u'compression': u"gzip[0-9]",
        u'cross_all_filesystems': u"0|1",
        u'cross_filesystem': u"path",
        u'daemon': u"0|1",
        u'dedup_group': u"string",
        u'directory_tree': u"0|1",
        u'directory': u"path",
        u'exclude_comp': u"extension",
        u'exclude_ext': u"extension",
        u'exclude_fs': u"fstype",
        u'exclude_regex': u"regular expression",
        u'exclude': u"path",
        u'fork': u"0|1",
        u'group': u"groupname",
        u'hard_quota': u"b/Kb/Mb/Gb",
        u'hardlinked_archive': u"0|1",
        u'include_ext': u"extension",
        u'include_regex': u"regular expression",
        u'include': u"path",
        u'include_glob': u"glob",
        u'keep': u"number",
        u'librsync': u"0|1",
        u'lockfile': u"path",
        u'manual_delete': u"path",
        u'max_children': u"number",
        u'max_file_size': u"b/Kb/Mb/Gb",
        u'max_hardlinks': u"number",
        u'max_status_children': u"number",
        u'max_storage_subdirs': u"number",
        u'min_file_size': u"b/Kb/Mb/Gb",
        u'network_timeout': u"s",
        u'nobackup': u"file name",
        u'notify_failure_arg': u"string",
        u'notify_failure_script': u"path",
        u'notify_success_arg': u"string",
        u'notify_success_changes_only': u"0|1",
        u'notify_success_script': u"path",
        u'notify_success_warnings_only': u"0|1",
        u'password': u"password",
        u'password_check': u"0|1",
        u'pidfile': u"path",
        u'port': u"port number",
        u'ratelimit': u"Mb/s",
        u'read_all_blockdevs': u"0|1",
        u'read_all_fifos': u"0|1",
        u'read_blockdev': u"path",
        u'read_fifo': u"path",
        u'restore_client': u"client",
        u'resume_partial': u"0|1",
        u'scan_problem_raises_error': u"0|1",
        u'server_script_arg': u"path",
        u'server_script_notify': u"0|1",
        u'server_script_post_arg': u"string",
        u'server_script_post_notify': u"0|1",
        u'server_script_post_run_on_fail': u"0|1",
        u'server_script_post': u"path",
        u'server_script_pre_arg': u"string",
        u'server_script_pre_notify': u"0|1",
        u'server_script_pre': u"path",
        u'server_script': u"path",
        u'soft_quota': u"b/Kb/Mb/Gb",
        u'ssl_cert_ca': u"path",
        u'ssl_cert_password': u"password",
        u'ssl_cert': u"path",
        u'ssl_ciphers': u"cipher list",
        u'ssl_dhfile': u"path",
        u'ssl_key_password': u"password",
        u'ssl_key': u"path",
        u'ssl_peer_cn': u"string",
        u'status_port': u"port number",
        u'stdout': u"0|1",
        u'strip_vss': u"0|1",
        u'suplit_vss': u"0|1",
        u'syslog': u"0|1",
        u'timer_arg': u"string",
        u'timer_script': u"path",
        u'timestamp_format': u"strftime format",
        u'umask': u"umask",
        u'user': u"username",
        u'version_warn': u"0|1",
        u'vss_drives': u"list of drive letters",
        u'working_dir_recovery_method': u"resume|use|delete",
    }
    values = {
        u'compression': [u'gzip{0}'.format(x) for x in range(0, 10)],
        u'mode': [u'client', u'server'],
        u'ssl_compression': [u'zlib{0}'.format(x) for x in range(0, 10)],
        u'status_address': [u'127.0.0.1', u'::1'],  # 127.0.0.1 / ::1
        u'working_dir_recovery_method': [u'use', u'delete', u'resume'],
    }
    files = [
        u'ca_burp_ca',
        u'ca_conf',
        u'notify_failure_script',
        u'notify_success_script',
        u'server_script_post',
        u'server_script_pre',
        u'server_script',
        u'ssl_cert_ca',
        u'ssl_cert',
        u'ssl_dhfile',
        u'ssl_key',
        u'timer_script',
    ]
    multi_srv = [
        u'exclude_comp',
        u'exclude_ext',
        u'exclude_fs',
        u'exclude_regex',
        u'exclude',
        u'include_ext',
        u'include_glob',
        u'include_regex',
        u'include',
        u'keep',
        u'notify_failure_arg',
        u'notify_success_arg',
        u'restore_client',
        u'server_script_arg',
        u'server_script_post_arg',
        u'server_script_pre_arg',
        u'timer_arg',
    ]
    boolean_srv = [
        u'atime',
        u'client_can_delete',
        u'client_can_diff',
        u'client_can_force_backup',
        u'client_can_list',
        u'client_can_restore',
        u'client_can_verify',
        u'cross_all_filesystems',
        u'daemon',
        u'directory_tree',
        u'fork',
        u'hardlinked_archive',
        u'librsync',
        u'monitor_browse_cache',
        u'notify_success_changes_only',
        u'notify_success_warnings_only',
        u'password_check',
        u'path_length_warn',
        u'read_all_blockdevs',
        u'read_all_fifos',
        u'scan_problem_raises_error',
        u'server_script_notify',
        u'server_script_post_notify',
        u'server_script_post_run_on_fail',
        u'server_script_pre_notify',
        u'split_vss',
        u'stdout',
        u'strip_vss',
        u'syslog',
        u'version_warn',
        u'server_can_restore'
    ]
    integer_srv = [
        u'max_children',
        u'max_hardlinks',
        u'max_status_children',
        u'max_storage_subdirs',
        u'network_timeout',
        u'port',
        u'protocol',
        u'ratelimit',
        u'status_port',
    ]
    string_srv = [
        u'address',
        u'ca_burp_ca',
        u'ca_conf',
        u'ca_name',
        u'ca_server_name',
        u'client_lockdir',
        u'compression',
        u'dedup_group',
        u'directory',
        u'group',
        u'hard_quota',
        u'mode',
        u'notify_failure_script',
        u'notify_success_script',
        u'pidfile',
        u'server_script_post',
        u'server_script_pre',
        u'server_script',
        u'soft_quota',
        u'ssl_cert_ca',
        u'ssl_cert',
        u'ssl_ciphers',
        u'ssl_compression',
        u'ssl_dhfile',
        u'ssl_key_password',
        u'ssl_key',
        u'status_address',
        u'timestamp_format',
        u'umask',
        u'user',
        u'working_dir_recovery_method',
        u'min_file_size',
        u'max_file_size',
        u'cross_filesystem',
        u'nobackup',
        u'read_fifo',
        u'read_blockdev',
        u'vss_drives',
    ]
    fields_cli = [
        u'atime',
        u'client_can_delete',
        u'client_can_force_backup',
        u'client_can_list',
        u'client_can_restore',
        u'client_can_verify',
        u'compression',
        u'cross_all_filesystems',
        u'cross_filesystem',
        u'dedup_group',
        u'directory_tree',
        u'directory',
        u'exclude_comp',
        u'exclude_ext',
        u'exclude_fs',
        u'exclude_regex',
        u'exclude',
        u'hard_quota',
        u'include_ext',
        u'include_regex',
        u'include',
        u'keep',
        u'librsync',
        u'max_file_size',
        u'min_file_size',
        u'nobackup',
        u'notify_failure_arg',
        u'notify_failure_script'
        u'notify_success_arg',
        u'notify_success_script',
        u'notify_success_warnings_only',
        u'password_check',
        u'password',
        u'path_length_warn',
        u'protocol',
        u'read_all_blockdevs',
        u'read_all_fifos',
        u'read_blockdev',
        u'read_fifo',
        u'restore_client',
        u'scan_problem_raises_error',
        u'server_script_arg',
        u'server_script_notify',
        u'server_script_post_arg',
        u'server_script_post_notify',
        u'server_script_post_run_on_fail',
        u'server_script_post',
        u'server_script_pre_arg',
        u'server_script_pre_notify',
        u'server_script_pre',
        u'server_script',
        u'soft_quota',
        u'split_vss',
        u'ssl_peer_cn',
        u'strip_vss',
        u'syslog'
        u'timer_arg',
        u'timer_script',
        u'timestamp_format',
        u'version_warn',
        u'vss_drives',
        u'working_dir_recovery_method',
        u'server_can_restore'
    ]
    string_cli = list(set(string_srv) & set(fields_cli))
    string_cli += [u'ssl_peer_cn', u'password']
    boolean_cli = list(set(boolean_srv) & set(fields_cli))
    integer_cli = list(set(integer_srv) & set(fields_cli))
    multi_cli = list(set(multi_srv) & set(fields_cli))
    doc = {
        u'.': u"Read additional configuration files. On Windows, the glob is"
              " unimplemented - you will need to specify an actual file.",
        u'address': u"Defines the main TCP address that the server listens"
                    " on. The default is either '::' or '0.0.0.0', dependent"
                    " upon compile time options.",
        u'atime': u"This allows you to control whether the client uses"
                  " O_NOATIME when opening files and directories. The"
                  " default is 0, which enables O_NOATIME. This means that"
                  " the client can read files and directories without"
                  " updating the access times. However, this is only"
                  " possible if you are running as root, or are the owner of"
                  " the file or directory. If this is not the case (perhaps"
                  " you only have group or world access to the files), you"
                  " will get errors until you set atime=1. With atime=1, the"
                  " access times will be updated on the files and"
                  " directories that get backed up.",
        u'autoupgrade_dir': u"Path to autoupgrade directory from which"
                            " upgrades are downloaded. The option can be"
                            " left unset in order not to autoupgrade"
                            " clients. Please see docs/autoupgrade.txt in"
                            " the source package for more help with this"
                            " option.",
        u'ca_burp_ca': u"Path to the burp_ca script when using the ca_conf"
                       " option.",
        u'ca_conf': u"Path to certificate authority configuration file. The"
                    " CA configuration file will usually be"
                    " /etc/burp/CA.cnf. The CA directory indicated by CA.cnf"
                    " will usually be /etc/burp/CA. If ca_conf is set and"
                    " the CA directory does not exist, the server will"
                    " create, populate it, and the paths indicated by"
                    " ssl_cert_ca, ssl_cert, ssl_key and ssl_dhfile will be"
                    " overwritten. For more detailed information on this and"
                    " the other ca_* options, please see docs/burp_ca.txt.",
        u'ca_name': u"Name of the CA that the server will generate when"
                    " using the ca_conf option.",
        u'ca_server_name': u"The name that the server will put into its own"
                           " SSL certficates when using the ca_conf option.",
        u'client_can_delete': u"Turn this off to prevent clients from"
                              " deleting backups with the '-a D' option. The"
                              " default is that clients can delete backups."
                              " Restore clients can override this setting.",
        u'client_can_force_backup': u"Turn this off to prevent clients from"
                                    " forcing backups with the '-a b'"
                                    " option. Timed backups will still work."
                                    " The default is that clients can force"
                                    " backups.",
        u'client_can_list': u"Turn this off to prevent clients from listing"
                            " backups with the '-a l' option. The default is"
                            " that clients can list backups. Restore clients"
                            " can override this setting.",
        u'client_can_restore': u"Turn this off to prevent clients from"
                               " initiating restores with the '-a r' option."
                               " The default is that clients can initiate"
                               " restores. Restore clients can override this"
                               " setting.",
        u'client_can_verify': u"Turn this off to prevent clients from"
                              " initiating a verify job with the '-a v'"
                              " option. The default is that clients can"
                              " initiate a verify job. Restore clients can"
                              " override this setting.",
        u'client_lockdir': u"Path to the directory in which to keep"
                           " per-client lock files. By default, this is set"
                           " to the path given by the 'directory' option.",
        u'clientconfdir': u"Path to the directory that contains client"
                          " configuration files.",
        u'compression': u"Choose the level of gzip compression for files"
                        " stored in backups. Setting 0 or gzip0 turns"
                        " compression off. The default is gzip9. This option"
                        " can be overridden by the client configuration"
                        " files in clientconfdir on the server.",
        u'cross_all_filesystems': u"Allow backups to cross all filesystem"
                                  " mountpoints.",
        u'cross_filesystem': u"Allow backups to cross a particular"
                             " filesystem mountpoint.",
        u'daemon': u"Whether to daemonise. The default is 1.",
        u'dedup_group': u"Enables you to group clients together for file"
                        " deduplication purposes. For example, you might"
                        " want to set 'dedup_group=xp' for each Windows XP"
                        " client, and then run the bedup program on a cron"
                        " job every other day with the option '-g xp'.",
        u'directory_tree': u"When turned on (which is the default) and the"
                           " client is on version 1.3.6 or greater, the"
                           " structure of the storage directory will mimic"
                           " that of the original filesystem on the client.",
        u'directory': u"Path to the directory in which to store backups.",
        u'exclude_comp': u"Extensions to exclude from compression. Case"
                         " insensitive. You can have multiple exclude"
                         " compression lines. For example, set 'gz' to"
                         " exclude gzipped files from compression.",
        u'exclude_ext': u"Extensions to exclude from the backup. Case"
                        " insensitive. You can have multiple exclude"
                        " extension lines. For example, set 'vdi' to exclude"
                        " VirtualBox disk images.",
        u'exclude_fs': u"File systems to exclude from the backup. Case"
                       " insensitive. You can have multiple exclude file"
                       " system lines. For example, set 'tmpfs' to exclude"
                       " tmpfs. Burp has an internal mapping of file system"
                       " names to file system IDs. If you know the file"
                       " system ID, you can use that instead. For example,"
                       " 'exclude_fs = 0x01021994' will also exclude tmpfs.",
        u'exclude_regex': u"Exclude paths that match the regular expression.",
        u'exclude': u"Path to exclude from the backup. You can have multiple"
                    " exclude lines. Use forward slashes '/', not"
                    " backslashes '\\' as path delimiters.",
        u'fork': u"Whether to fork children. The default is 1.",
        u'group': u"Run as a particular group. This can be overridden by the"
                  " client configuration files in clientconfdir on the"
                  " server.",
        u'hard_quota': u"Do not back up the client if the estimated size of"
                       " all files is greater than the specified size."
                       " Example: 'hard_quota = 100Gb'. Set to 0 (the"
                       " default) to have no limit.",
        u'hardlinked_archive': u"On the server, defines whether to keep"
                               " hardlinked files in the backups, or whether"
                               " to generate reverse deltas and delete the"
                               " original files. Can be set to either 0"
                               " (off) or 1 (on). Disadvantage: More disk"
                               " space will be used Advantage: Restores will"
                               " be faster, and since no reverse deltas need"
                               " to be generated, the time and effort the"
                               " server needs at the end of a backup is"
                               " reduced.",
        u'include_ext': u"Extensions to include in the backup. Case"
                        " insensitive. Nothing else will be included in the"
                        " backup. You can have multiple include extension"
                        " lines. For example, set 'txt' to include files"
                        " that end in '.txt'. You need to specify an"
                        " 'include' line so that burp knows where to start"
                        " looking.",
        u'include_regex': u"Not implemented.",
        u'include': u"Path to include in the backup. You can have multiple"
                    " include lines. Use forward slashes '/', not"
                    " backslashes '\\' as path delimiters.",
        u'include_glob': u"Include paths that match the glob expression. For"
                         " example, '/home/*/Documents' will include"
                         " '/home/user1/Documents' and"
                         " '/home/user2/Documents' if directories 'user1'"
                         " and 'user2' exist in '/home'. The Windows"
                         " implementation currently limit the expression to"
                         " contain only one '*'.",
        u'keep': u"Number of backups to keep. This can be overridden by the"
                 " client configuration files in clientconfdir on the"
                 " server. Specify multiple 'keep' entries on separate lines"
                 " in order to keep multiple periods of backups. For"
                 " example, assuming that you are doing a backup a day,"
                 " keep=7 keep=4 keep=6 (on separate lines) will keep 7"
                 " daily backups, 4 weekly backups (7x4=28), and 6 multiples"
                 " of 4 weeks (7x4x6=168) - roughly 6 monthly backups."
                 " Effectively, you will be guaranteed to be able to restore"
                 " up to 168 days ago, with the number of available backups"
                 " exponentially decreasing as you go back in time to that"
                 " point. In this example, every 7th backup will be"
                 " hardlinked to allow burp to safely delete intermediate"
                 " backups when necessary. You can have as many 'keep' lines"
                 " as you like, as long as they don't exceed 52560000 when"
                 " multiplied together. That is, a backup every minute for"
                 " 100 years.",
        u'librsync': u"When set to 0, delta differencing will not take"
                     " place. That is, when a file changes, the server will"
                     " request the whole new file. The default is 1. This"
                     " option can be overridden by the client configuration"
                     " files in clientconfdir on the server.",
        u'lockfile': u"Path to the lockfile that ensures that two server"
                     " processes cannot run simultaneously.",
        u'manual_delete': u"If a path is given, the server will move"
                          " directories to be deleted into the directory"
                          " specified by the path, but will not actually"
                          " delete them. The path must be on the same file"
                          " system as the backup storage. The idea is that a"
                          " busy server may be configured to run the"
                          " deletions outside of the backup timebands, when"
                          " the server is less busy, via a cron job. The"
                          " default is unset, which means that the server"
                          " will automatically delete the directories at the"
                          " end of a backup. This option can be overridden"
                          " by the client configuration files in"
                          " clientconfdir on the server.",
        u'max_children': u"Defines the number of child processes to fork"
                         " (the number of clients that can simultaneously"
                         " connect. The default is 5.",
        u'max_file_size': u"Do not back up files that are greater than the"
                          " specified size. Example: 'max_file_size = 10Mb'."
                          " Set to 0 (the default) to have no limit.",
        u'max_hardlinks': u"On the server, the number of times that a single"
                          " file can be hardlinked. The bedup program also"
                          " obeys this setting. The default is 10000.",
        u'max_status_children': u"Defines the number of status child"
                                " processes to fork (the number of status"
                                " clients that can simultaneously connect."
                                " The default is 5.",
        u'max_storage_subdirs': u"Defines the number of subdirectories in"
                                " the data storage areas. The maximum number"
                                " of subdirectories that ext3 allows is"
                                " 32000. If you do not set this option, it"
                                " defaults to 30000.",
        u'min_file_size': u"Do not back up files that are less than the"
                          " specified size. Example: 'min_file_size = 10Mb'."
                          " Set to 0 (the default) to have no limit.",
        u'mode': u"Required to run in server mode.",
        u'monitor_browse_cache': u"Whether or not the server should cache"
                                 " the directory tree when a monitor client"
                                 " is browsing. <br/>Advantage: browsing is"
                                 " faster. </br>Disadvantage: more memory is"
                                 " used.",
        u'network_timeout': u"Set the network timeout in seconds. If no data"
                            " is sent or received over a period of this"
                            " length, burp will give up. The default is 7200"
                            " seconds (2 hours).",
        u'nobackup': u"If this file system entry exists, the directory"
                     " containing it will not be backed up.",
        u'notify_failure_arg': u"The same as notify_success_arg, but for"
                               " backups that failed.",
        u'notify_failure_script': u"The same as notify_success_script, but"
                                  " for backups that failed.",
        u'notify_success_arg': u"A user-definable argument to the notify"
                               " success script. You can have many of these."
                               " The notify_success_arg options can be"
                               " overridden by the client configuration"
                               " files in clientconfdir on the server.",
        u'notify_success_script': u"Path to the script to run when a backup"
                                  " succeeds. User arguments are appended"
                                  " after the first five reserved arguments."
                                  " An example notify script is provided."
                                  " The notify_success_script option can be"
                                  " overridden by the client configuration"
                                  " files in clientconfdir on the server.",
        u'notify_success_warnings_only': u"Set to 1 to send success"
                                         " notifications when there were"
                                         " warnings. If this and"
                                         " notify_success_changes_only are"
                                         " not turned on, success"
                                         " notifications are always sent.",
        u'password': u"Defines the password to send to the server.",
        u'password_check': u"Allows you to turn client password checking on"
                           " or off. The default is on. SSL certificates"
                           " will still be checked if you turn passwords"
                           " off. This option can be overridden by the"
                           " client configuration files in clientconfdir on"
                           " the server.",
        u'path_length_warn': u"When this is on, which is the default, a"
                             " warning will be issued when the client sends"
                             " a path that is too long to replicate in the"
                             " storage area tree structure. The file will"
                             " still be saved in a numbered file outside of"
                             " the tree structure, regardless of the setting"
                             " of this option. This option can be overridden"
                             " by the client configuration files in"
                             " clientconfdir on the server.",
        u'pidfile': u"Synonym for lockfile.",
        u'port': u"Defines the main TCP port that the server listens on.",
        u'protocol': u"Choose which style of backups and restores to use. 0"
                     " (the default) automatically decides based on the"
                     " server version and which protocol is set on the"
                     " server side. 1 forces protocol1 style (file level"
                     " granularity with a pseudo mirrored storage on the"
                     " server and optional rsync). 2 forces protocol2 style"
                     " (inline deduplication with variable length blocks)."
                     " If you choose a forced setting, it will be an error"
                     " if the server also chooses a forced setting.",
        u'ratelimit': u"Set the network send rate limit, in Mb/s. If this"
                      " option is not given, burp will send data as fast as"
                      " it can.",
        u'read_all_blockdevs': u"Open all block devices for reading and back"
                               " up the contents as if they were regular"
                               " files.",
        u'read_all_fifos': u"Open all fifos for reading and back up the"
                           " contents as if they were regular files.",
        u'read_blockdev': u"Do not back up the given block device itself,"
                          " but open it for reading and back up the contents"
                          " as if it were a regular file.",
        u'read_fifo': u"Do not back up the given fifo itself, but open it"
                      " for reading and back up the contents as if it were a"
                      " regular file.",
        u'restore_client': u"A client that is permitted to list, verify,"
                           " restore and delete files belonging to any other"
                           " client. You may specify multiple"
                           " restore_clients. If this is too permissive, you"
                           " may set a restore_client for individual"
                           " original clients in the individual"
                           " clientconfdir files. Note that restoring a"
                           " backup from a Windows computer onto a Linux"
                           " computer will currently leave the VSS headers"
                           " in place at the beginning of each file. This"
                           " will be addressed in a future version of burp.",
        u'resume_partial': u"Turn this on to enable 'resume partial' code."
                           " Requires 'working_dir_recovery_method=resume'."
                           " When resuming an interrupted transfer of a"
                           " single file, it attempts to use previously"
                           " transferred blocks of that file in order to be"
                           " more efficient. However, situations have been"
                           " reported where the file on the server side just"
                           " gets bigger forever, so this feature now"
                           " defaults to being turned off.",
        u'scan_problem_raises_error': u"When enabled, this causes problems"
                                      " in the phase1 scan (such as an"
                                      " 'include' being missing) to be"
                                      " treated as fatal errors. The default"
                                      " is off.",
        u'server_script_arg': u"Goes with server_script and overrides"
                              " server_script_pre_arg and"
                              " server_script_post_arg.",
        u'server_script_notify': u"Turn on to send a notification emails"
                                 " when the server pre and post scripts"
                                 " return non-zero. The output of the script"
                                 " will be included it the email. The"
                                 " default is off. Requires the"
                                 " notify_failure options to be set.",
        u'server_script_post_arg': u"A user-definable argument to the server"
                                   " post script. You can have many of these.",
        u'server_script_post_notify': u"Turn on to send a notification email"
                                      " when the server post script returns"
                                      " non-zero. The output of the script"
                                      " will be included in the email. The"
                                      " default is off. Requires the"
                                      " notify_failure options to be set.",
        u'server_script_post_run_on_fail': u"If this is set to 1,"
                                           " server_script_post will always"
                                           " be run. The default is 0, which"
                                           " means that if the task asked"
                                           " for by the client fails,"
                                           " server_script_post will not be"
                                           " run.",
        u'server_script_post': u"Path to a script to run on the server"
                               " before the client disconnects. The"
                               " arguments to it are 'post', '(client"
                               " command)', 'reserved3' to 'reserved5', and"
                               " then arguments defined by"
                               " server_script_post_arg. This command and"
                               " related options can be overriddden by the"
                               " client configuration files in clientconfdir"
                               " on the server.",
        u'server_script_pre_arg': u"A user-definable argument to the server"
                                  " pre script. You can have many of these.",
        u'server_script_pre_notify': u"Turn on to send a notification email"
                                     " when the server pre script returns"
                                     " non-zero. The output of the script"
                                     " will be included in the email. The"
                                     " default is off. Most people will not"
                                     " want this turned on because clients"
                                     " usually contact the server at 20"
                                     " minute intervals and this could cause"
                                     " a lot of emails to be generated."
                                     " Requires the notify_failure options"
                                     " to be set.",
        u'server_script_pre': u"Path to a script to run on the server after"
                              " each successfully authenticated connection"
                              " but before any work is carried out. The"
                              " arguments to it are 'pre', '(client"
                              " command)', 'reserved3' to 'reserved5', and"
                              " then arguments defined by"
                              " server_script_pre_arg. If the script returns"
                              " non-zero, the task asked for by the client"
                              " will not be run. This command and related"
                              " options can be overriddden by the client"
                              " configuration files in clientconfdir on the"
                              " server.",
        u'server_script': u"You can use this to save space in your config"
                          " file when you want to run the same server script"
                          " twice. It overrides server_script_pre and"
                          " server_script_post. This command and related"
                          " options can be overriddden by the client"
                          " configuration files in clientconfdir on the"
                          " server.",
        u'soft_quota': u"A warning will be issued when the estimated size of"
                       " all files is greater than the specified size and"
                       " smaller than hard_quota. Example: 'soft_quota ="
                       " 95Gb'. Set to 0 (the default) to have no warning.",
        u'split_vss': u"When backing up Windows computers with burp protocol"
                      " 1, this option allows you to save the VSS header"
                      " data separate from the file data. The default is"
                      " off, which means that the VSS header data is saved"
                      " prepended to the file data.",
        u'ssl_cert_ca': u"The path to the SSL CA certificate. This file will"
                        " probably be the same on both the server and the"
                        " client. The file should contain just the"
                        " certificate in PEM format. For more information on"
                        " this, and the other ssl_* options, please see"
                        " <a href='http://burp.grke.org/docs/burp_ca.html'>"
                        " docs/burp_ca.txt</a>.",
        u'ssl_cert_password': u"Synonym for ssl_key_password.",
        u'ssl_cert': u"The path to the server SSL certificate. It works for"
                     " me when the file contains the concatenation of the"
                     " certificate and private key in PEM format.",
        u'ssl_ciphers': u"Allowed SSL ciphers. See openssl ciphers for"
                        " details.",
        u'ssl_compression': u"Choose the level of zlib compression over SSL."
                            " Setting 0 or zlib0 turnsSSL compression off."
                            " Setting non-zero gives zlib5 compression (it"
                            " is not currently possible for openssl to set"
                            " any other level). The default is 5. 'gzip' is"
                            " a synonym of 'zlib'.is a synonym of 'zlib'.",
        u'ssl_dhfile': u"Path to Diffie-Hellman parameter file. To generate"
                       " one with openssl, use a command like this: openssl"
                       " dhparam -out dhfile.pem -5 1024",
        u'ssl_key_password': u"The SSL key password.",
        u'ssl_key': u"The path to the server SSL private key in PEM format.",
        u'status_address': u"Defines the main TCP address that the server"
                           " listens on for status requests. The default is"
                           " either '::1' or '127.0.0.1', dependent upon"
                           " compile time options.",
        u'status_port': u"Defines the TCP port that the server listens on"
                        " for status requests.",
        u'stdout': u"Log to stdout. Defaults to on.",
        u'strip_vss': u"When backing up Windows computers with burp protocol"
                      " 1, this option allows you to prevent the VSS header"
                      " data being backed up. The default is off. To restore"
                      " a backup that has no VSS information on Windows, you"
                      " need to give the client the '-x' command line option.",
        u'syslog': u"Log to syslog. Defaults to off.",
        u'timer_arg': u"A user-definable argument to the timer script. You"
                      " can have many of these. The timer_arg options can be"
                      " overridden by the client configuration files in"
                      " clientconfdir on the server.",
        u'timer_script': u"Path to the script to run when a client connects"
                         " with the timed backup option. If the script"
                         " exits with code 0, a backup will run. The first"
                         " two arguments are the client name and the path"
                         " to the 'current' storage directory. The next"
                         " three arguments are reserved, and user arguments"
                         " are appended after that. An example timer script"
                         " is provided. The timer_script option can be"
                         " overridden by the client configuration files in"
                         " clientconfdir on the server.",
        u'timestamp_format': u"This allows you to tweak the format of the"
                             " timestamps of individual backups. See 'man"
                             " strftime' to see available substitutions."
                             " If this option is unset, burp uses"
                             " \"%Y-%m-%d %H:%M:%S\".",
        u'umask': u"Set the file creation umask. Default is 0022.",
        u'user': u"Run as a particular user. This can be overridden by the"
                 " client configuration files in clientconfdir on the server.",
        u'version_warn': u"When this is on, which is the default, a warning"
                         " will be issued when the client version does not"
                         " match the server version. This option can be"
                         " overridden by the client configuration files in"
                         " clientconfdir on the server.",
        u'vss_drives': u"When backing up Windows computers, this option"
                       " allows you to specify which drives have VSS"
                       " snapshots taken of them. If you omit this option,"
                       " burp will automatically decide based on the"
                       " 'include' options. If you want no drives to have"
                       " snapshots taken of them, you can specify '0'.",
        u'working_dir_recovery_method': u"This option tells the server what"
                                        " to do when it finds the working"
                                        " directory of an interrupted backup"
                                        " (perhaps somebody pulled the plug"
                                        " on the server, or something). This"
                                        " can be overridden by the client"
                                        " configurations files in"
                                        " clientconfdir on the server."
                                        " Options are... <ul><li>delete:"
                                        " Just delete the old working"
                                        " directory.</li><li>use: Convert"
                                        " the working directory into a"
                                        " complete backup.</li><li>resume:"
                                        " Simply continue the previous"
                                        " backup from the point at which it"
                                        " left off, at file granularity."
                                        " NOTE: If the client has changed"
                                        " its include/exclude configuration"
                                        " since the backup was interrupted,"
                                        " the recovery method will"
                                        " automatically switch to 'use'."
                                        " </li></ul>",
    }
